import os
import frag_pele.constants as cs


def check():
    if not os.path.exists(cs.SCHRODINGER_PY_PATH):
        raise OSError("Schrodinger path {} not found. Change the harcoded path under frag_pele/constants.py".format(cs.SCHRODINGER_PY_PATH))
    if not os.path.exists(cs.PATH_TO_PELE):
        raise OSError("Pele path {} not found. Change the harcoded path under frag_pele/constants.py".format(cs.PATH_TO_PELE))
    elif not os.path.exists(cs.PATH_TO_PELE_DATA):
        raise OSError("Pele data path {} not found. Change the harcoded path under frag_pele/constants.py".format(cs.PATH_TO_PELE_DATA))
    elif not os.path.exists(cs.PATH_TO_PELE_DOCUMENTS):
        raise OSError("Pele documents path {} not found. Change the harcoded path under frag_pele/constants.py".format(cs.PATH_TO_PELE_DOCUMENTS))
    elif not os.path.exists(cs.PATH_TO_LICENSE):
        raise OSError("Pele license path {} not found. Change the harcoded path under frag_pele/constants.py".format(cs.PATH_TO_LICENSE))



def main(pdb_path, lig_chain="L"):
    with open(pdb_path) as in_pdb:
        pdb = in_pdb.readlines()
    elements = []
    pdb_out = []
    dictionary_to_transcript = {}
    for line in pdb:
        if line[0:6] == "HETATM" and line[21:22] == lig_chain:
            element = line[76:78]
            elements.append(element)
            list_of_unique_elements = set(elements)
            counters = {i: elements.count(i) for i in list_of_unique_elements}
            counter = counters[line[76:78]]
            new_name = "{}{}".format(line[76:78].strip().upper(), counter)
            while True:
                new_name = new_name + " "
                if len(new_name) == 4:
                    break
            dictionary_to_transcript[line[12:16]] = new_name
            line = list(line)
            line[12:16] = new_name
            line = "".join(line)
            if len(new_name) > 4:
                raise ValueError("Length of the string {} is too long. Only 4 characters accepted.".format(new_name))
        pdb_out.append(line)

    return dictionary_to_transcript



import prody
#import pybel
import logging
import sys

# Getting the name of the module for the log system
logger = logging.getLogger(__name__)


def pdb_parser_ligand(pdb_file, ligand_chain="L"):
    """
    :param pdb_file: input PDB file of the complex that we want to get their ligand
    :param ligand_chain: chain where the ligand is placed
    :return: PRODY object with the atoms of the ligand of the input PDB
    """
    pdb = prody.parsePDB(pdb_file)
    ligand = pdb.select("chain {}".format(ligand_chain))
    if ligand is None:
        logger.critical("Wrong chain selected!")
    elif ligand.ishetero:
        return ligand
    else:
       logger.critical("The selected chain does not contain heteroatoms!")


def selection_to_pdb(selection):
    """
    :param selection: prody selection
    :return: writes a PDB file containing the selection, named using the residue name of the first atom of the selection.
    """
    prody.writePDB(selection.getResnames()[0], selection)
    return "{}.pdb".format(selection.getResnames()[0])


def check_protonation(selection):
    """
    Check if the structure is protonated or not. In case that is not protonated we will rise a critical logger.
    :param selection: prody molecule
    :return: if not hydrogens detected, prints a message.
    """
    try:
        if not selection.select("hydrogen"):
            logger.critical("We have not detected Hydrogens in your ligand. Please, add them before starting.")
    except AttributeError:
        raise AttributeError("Check ligand and core are in the L chain. Otherwise specify their chain using the flags \
-cc chain_core -fc chain_frag. Fragment & Core must always be in the same chain.")





import os
import sys
from schrodinger import structure as st
import subprocess
import argparse
# Local import
frag_pele_dirname = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
sys.path.append(frag_pele_dirname)
import frag_pele.PlopRotTemp_S_2017.main as plop


def convert_mae(ligands):
    """
       Desciption: From each structure retrieve
       a .mae file of the ligand in the receptor.
       Output:
            structure_mae: ligand
            res = residue
    """

    
    for structure in st.StructureReader(ligands):
        for residue in structure.residue:
            res = residue.pdbres.strip()
        str_name = "{}".format(res)
        try:
            structure.write(str_name + ".mae")
        except ValueError:
            str_name = "{}".format(res)
        finally:
            structure_mae = "{}.mae".format(str_name)
            structure.write(structure_mae)
    return structure_mae


def create_template(pdb, gridres, out_temp, out_lib):
   mae_file = convert_mae(pdb)
   plop.main(mae_file, out_temp=out_temp, out_rot=out_lib, gridres=gridres)
   os.remove(mae_file)


def arg_parse():
  parser = argparse.ArgumentParser()

  parser.add_argument("pdb", type=str, help="ligand file to templatize")
  parser.add_argument("gridres", type=str, help="Degrees of rotation.")
  parser.add_argument("out_temp", type=str, help="Output template path.")
  parser.add_argument("out_lib", type=str, help="Output rotamer library path.")

  args = parser.parse_args()

  return args.pdb, args.gridres, args.out_temp, args.out_lib


if __name__ == '__main__':
    pdb, gridres, outtemp, outlib = arg_parse()
    create_template(pdb, gridres, outtemp, outlib)



import prody
import sys
import logging
import numpy as np
import pickle
from scipy.spatial import distance
import math
import os
import shutil
import re
import Bio.PDB as bio
# Local imports
import frag_pele.constants as c
from frag_pele.Helpers import checker
from frag_pele.Growing.AddingFragHelpers import complex_to_prody, pdb_joiner, atom_constants
try:
    import rdkit
    RDKIT = True
except ImportError:
    RDKIT = False
if RDKIT:
    from lib_prep.FragmentTools import tree_detector

# Getting the name of the module for the log system
logger = logging.getLogger(__name__)
# Lists
LIST_OF_IONS = ["ZN", "MN", "FE", "CO", "NI", "CA", "CD"]


def extract_heteroatoms_pdbs(pdb, create_file=True, ligand_chain="L", get_ligand=False, output_folder="."):
    """
    From a pdb file, it extracts the chain L and checks if the structure has hydrogens. After that, the chain L is
    written in a new PDB file which will have the following format: "{residue name}.pdb".
    :param pdb: pdb file (with a ligand in the chain L).
    :return: Writes a new pdb file "{residue name}.pdb" with the chain L isolated an returns the residue name (string).
    """
    # Parse the complex file and isolate the ligand core and the fragment
    ligand = complex_to_prody.pdb_parser_ligand(pdb, ligand_chain)
    if ligand is None:
        logger.critical("The ligand can not be found. Ensure that the ligand of {} is the chain {}".format(pdb, ligand_chain))
    # Check if the ligand has H
    complex_to_prody.check_protonation(ligand)
    # Save the ligand in a PDB (the name of the file is the name of the residue)
    ligand_name = ligand.getResnames()[0]
    if create_file:
        prody.writePDB(os.path.join(output_folder, ligand_name), ligand)
        print("The ligand of {} has been extracted and saved in '{}.pdb'".format(pdb, os.path.join(output_folder, ligand_name)))
    if get_ligand is True:
        return ligand
    else:
        return ligand_name


def from_pdb_to_bioatomlist(list_of_pdb_names):
    """
    Given a pdb name string (without the extension ".pdb") the function reads it as a Bio.PDB structure and extract the
    atoms found as a list of Bio.PDB.Atom objects.
    :param list_of_pdb_names: list of strings with pdb names.
    :return: list of lists with the Bio.PDB.Atom objects found in each pdb file.
    """
    list_of_lists = []
    for pdb in list_of_pdb_names:
        # Now, read the PDBs as BioPython object
        bio_structure = pdb_joiner.get_ligand_from_PDB("{}.pdb".format(pdb))
        # Get a list of atoms (BioPython Atoms) for each structure
        bioatomlist = pdb_joiner.get_atoms_from_structure(bio_structure)
        list_of_lists.append(bioatomlist)
    return list_of_lists


def extract_heavy_atoms(pdb_atom_names, lists_of_bioatoms):
    """
    Given a heavy atom name (string) and a list of Bio.PDB.Atom objects, it selects this atom of the list and return it
    as a single object.
    :param pdb_atom_names: heavy atom name (string).
    :param lists_of_bioatoms: list of Bio.PDB.Atom objects.
    :return: Bio.PDB.Atom object correspondent to the heavy atom name.
    """
    heavy_atoms = []
    # Select the heavy atoms for each list that we will want to bond together in further steps
    for atom_name, list_of_bioatoms in zip(pdb_atom_names, lists_of_bioatoms):
        atom_heavy = pdb_joiner.select_atoms_from_list(atom_name, list_of_bioatoms)
        heavy_atoms.append(atom_heavy)
    return heavy_atoms


def extract_hydrogens(pdb_atom_names, lists_of_bioatoms, list_of_pdbs, h_core=None, h_frag=None, c_chain="L", f_chain="L"):
    """
    Given a heavy atom name (string), a list of Bio.PDB.Atoms objects and a list of pdb files, it returns the hydrogens
    at bonding distance of the heavy atom. If there is more than one, a checking of contacts with the
    protein will be performed. In case of finding a possible contact between the hydrogen and the protein, we will
    reject this hydrogen and we will repeat the check in another one. If all the hydrogens have contacts with the
    protein, the first of them will be selected and a warning will be printed.
    :param pdb_atom_names: heavy atom name (string).
    :param lists_of_bioatoms: list of Bio.PDB.Atom objects.
    :param list_of_pdbs: list of PDB files.
    :return: Bio.PDB.Atom object correspondent to the hydrogen bonded to the heavy atom
    """

    hydrogens = []
    selected_hydrogens = [h_core, h_frag]
    chains = [c_chain, f_chain]
    for atom_name, pdb, list_of_bioatoms, sel_h, chain in zip(pdb_atom_names, list_of_pdbs, lists_of_bioatoms, selected_hydrogens, chains):
        complex = prody.parsePDB(pdb)
        # Select name of the H atoms bonded to this heavy atom (the place where we will grow)
        atom_name_hydrogens = pdb_joiner.get_H_bonded_to_grow(atom_name, complex, sel_h, chain=chain)
        # Select this hydrogen atoms
        atom_hydrogen = pdb_joiner.select_atoms_from_list(atom_name_hydrogens, list_of_bioatoms)
        hydrogens.append(atom_hydrogen)
    return hydrogens


def transform_coords_from_bio2prody(fragment_prody, bioatom_list):
    """
    Given a fragment (prody molecule object) and a list of Bio.PDB.Atom objects correspondent to this fragment, it
    replace the coordinates of the prody molecule for the ones of the list of Bio.PDB.Atom objects and returns the new
    coordinates.
    :param fragment_prody: prody molecule object.
    :param bioatom_list: list of Bio.PDB.Atom objects.
    :return: array with the new coordinates of the prody molecule object.
    """
    fragment_coords = pdb_joiner.transform_coords(bioatom_list)
    fragment_prody.setCoords(fragment_coords)
    return fragment_prody.getCoords()


def bond(hydrogen_atom_names, molecules):
    """
    Given a list with names of hydrogens (bonded to the heavy atoms that we want to link) and a list of molecules (prody
    molecule object), this function errase this hydrogens and bond the heavy atoms linked to them. In order to create
    this new bond we concatenate the pairs of molecules (prody molecule object).
    Note that the lists need to have pairs of hydrogens and molecules because the bonding is going to happen between
    adjacent pairs of hydrogens atoms and molecules.
    :param hydrogen_atom_names: names of hydrogens (bonded to the heavy atoms that we want to link). Each two elements
    of the list will be considered as hydrogens that will form the bond.
    :param molecules: prody molecule object. Each two elements of the list will be considered as molecules
    that will be bonded.
    :return: list of prody molecule objects as a result of merging the pairs of molecules.
    """
    list_of_pairs = []
    for hydrogen, molecule in zip(hydrogen_atom_names, molecules):
        # Now, we have to remove the hydrogens of the binding, so we will select everything except these H
        mol_no_h = molecule.select("not name {}".format(hydrogen))
        list_of_pairs.append(mol_no_h)
    # Merging both parts into a single one
    i = 0
    bonds = []
    while i < len(list_of_pairs):
        merged = list_of_pairs[i].copy() + list_of_pairs[i+1].copy()
        bonds.append(merged)
        i += 2
    return bonds

def join_structures_to_rotate(core_bond, fragment_bond, list_of_atoms, core_structure, fragment_structure):
    """
    It joins two ProDy structures into a single one, merging both bonds (core bond and fragment bond) creating a unique
    bond between the molecules. In order to do that this function performs a cross superimposition (in BioPython) of
    the whole fragment using as reference (fixed part) the atoms of the bond. Then, it transforms this BioPython object
    into ProDy molecule with the coordinates modified. Once we have all ready, the Hydrogens of the bonds will be
    deleted and both structures will be concatenated into a single ProDy object.
    :param core_bond: Bio.PDB.Atom list with two elements: [heavy atom, hydrogen atom]. These two atoms have to be the
    ones participating in the bond of the core that we would like to use as linking point between core and fragment.
    :param fragment_bond: Bio.PDB.Atom list with two elements: [hydrogen atom, heavy atom]. These two atoms have to be the
    ones participating in the bond of the fragment that we would like to use as linking point between fragment and core.
    :param list_of_atoms: Bio.PDB.Atom list with all the atoms of the fragment. These atoms have to contain
    coordinates.
    :param core_structure: ProDy molecule that contain only the core ligand.
    :param fragment_structure: ProDy molecule that contain only the fragment ligand.
    :return: ProDy molecule with the core_structure and the fragment_structure (with the coordinates modified)
    concatenated.
    """
    # Superimpose atoms of the fragment to the core bond
    pdb_joiner.superimpose(core_bond, fragment_bond, list_of_atoms)
    # Get the new coords and change them in prody
    transform_coords_from_bio2prody(fragment_structure, list_of_atoms)
    # Now, we have to remove the hydrogens of the binding
    h_atom_names = [core_bond[1].name, fragment_bond[0].name]
    merged_structure = bond(h_atom_names, [core_structure, fragment_structure])
    return merged_structure


def join_structures(core_bond, fragment_bond, core_structure, fragment_structure, pdb_complex,
                    pdb_fragment, chain_complex, chain_fragment, output_path, only_grow=False):
    """
    It joins two ProDy structures into a single one, merging both bonds (core bond and fragment bond) creating a unique
    bond between the molecules. In order to do that this function performs a cross superimposition (in BioPython) of
    the whole fragment using as reference (fixed part) the atoms of the bond. Then, it transforms this BioPython object
    into ProDy molecule with the coordinates modified. Once we have all ready, the Hydrogens of the bonds will be
    deleted accordingly with the bond type, and both structures will be concatenated into a single ProDy object.

    :param core_bond: Bio.PDB.Atom list with two elements: [heavy atom, hydrogen atom]. These two atoms have to be the
    ones participating in the bond of the core that we would like to use as linking point between core and fragment.
    :param fragment_bond: Bio.PDB.Atom list with two elements: [hydrogen atom, heavy atom]. These two atoms have to be the
    ones participating in the bond of the fragment that we would like to use as linking point between fragment and core.
    :param list_of_atoms: Bio.PDB.Atom list with all the atoms of the fragment. These atoms have to contain
    coordinates.
    :param core_structure: ProDy molecule that contain only the core ligand.
    :param fragment_structure: ProDy molecule that contain only the fragment ligand.
    :param pdb_complex: path to the file in PDB with the protein-ligand complex.
    :param pdb_fragment: path to the file in PDB with the fragment.
    :param chain_complex: label of the ligand chain in the pdb_complex.
    :param chain_fragment: label of the ligand chain in the pdb_fragment.
    :return: ProDy molecule with the core_structure and the fragment_structure (with the coordinates modified)
    concatenated.
    """
    name_to_replace_core = core_bond[1].name
    name_to_replace_fragment = fragment_bond[0].name
    if only_grow:
        return 0, name_to_replace_core, name_to_replace_fragment
    if RDKIT:
        atoms_to_delete_core = tree_detector.main(pdb_complex, (core_bond[0].name, name_to_replace_core),
                                                  chain_ligand=chain_complex)
        atoms_to_delete_fragment = tree_detector.main(pdb_fragment, (fragment_bond[1].name, name_to_replace_fragment),
                                                      chain_ligand=chain_fragment)
    else:
        print("WARNING: YOU CAN NOT REPLACE HEAVY ATOMS FOR HYDROGENS WITHOUT RDKIT!")
    fragment_rdkit = rdkit.Chem.MolFromPDBFile(pdb_fragment, removeHs=False)
    bond_type = detect_bond_type(fragment_rdkit, fragment_bond[0], fragment_bond[1])
    if RDKIT:
        if core_bond[1].element != "H":
            atom_replaced_idx = replace_heavy_by_hydrogen(core_bond[1], core_structure)
            new_coords, new_dist = correct_hydrogen_position(hydrogen_atom=core_structure[atom_replaced_idx],
                                                             atom_to_bond_with=core_bond[0],
                                                             structure=core_structure[atom_replaced_idx])
            core_structure[atom_replaced_idx].setCoords(new_coords)
            names_to_keep = list(
                set(core_structure.getNames()) ^ set(atoms_to_delete_core))  # Compare two sets and get the common items
            names_to_keep.remove(name_to_replace_core)
            core_structure = core_structure.select("name {}".format(" ".join(names_to_keep)))
            prody.writePDB(os.path.join(output_path, "{}.pdb".format(core_structure.getResnames()[0])),
                           core_structure)  # Overwrite the initial structure
            name_to_replace_core = core_structure[atom_replaced_idx].getName()
            core_bond[1].coord = new_coords
        if fragment_bond[0].element != "H":
            atom_replaced_idx = replace_heavy_by_hydrogen(fragment_bond[0], fragment_structure)
            new_coords, new_dist = correct_hydrogen_position(hydrogen_atom=fragment_structure[atom_replaced_idx],
                                                             atom_to_bond_with=fragment_bond[1],
                                                             structure=fragment_structure[atom_replaced_idx])
            fragment_structure[atom_replaced_idx].setCoords(new_coords)
            names_to_keep = list(
                set(fragment_structure.getNames()) ^ set(atoms_to_delete_fragment))  # Compare two sets and get the common items
            names_to_keep.remove(name_to_replace_fragment)
            fragment_structure = fragment_structure.select("name {}".format(" ".join(names_to_keep)))
            prody.writePDB(os.path.join(output_path, "{}.pdb".format(fragment_structure.getResnames()[0])),
                           fragment_structure)  # Overwrite the initial structure
            name_to_replace_fragment = fragment_structure[atom_replaced_idx].getName()
            fragment_bond[0].coord = new_coords
    bio_list = from_pdb_to_bioatomlist([os.path.join(output_path, "{}".format(fragment_structure.getResnames()[0]))])[0] # Its a list, so we keep only the unique element that is inside
    # Superimpose atoms of the fragment to the core bond
    pdb_joiner.superimpose(core_bond, fragment_bond, bio_list)
    # Get the new coords and change them in prody
    transform_coords_from_bio2prody(fragment_structure, bio_list)
    # Now, we have to remove the hydrogens of the binding
    h_atom_names = [name_to_replace_core, name_to_replace_fragment]
    # Correcting linking distance
    new_coords, new_distance = correct_bonding_distance(atom_reference=core_bond[0], atom_to_correct=fragment_bond[1],
                                                        reference_structure=core_structure, movil_structure=fragment_structure,
                                                        bond_type=bond_type)
    if bond_type == "double":
        hydrogen_to_delete = pdb_joiner.get_H_bonded_to_atom(core_bond[0].name, core_structure,
                                                             bond_dist=new_distance,
                                                 	     banned_hydrogen=name_to_replace_core,
                                                             chain=chain_complex)
        names_to_keep = list(core_structure.getNames())
        names_to_keep.remove(hydrogen_to_delete)
        core_structure = core_structure.select("name {}".format(" ".join(names_to_keep)))

    if bond_type == "triple":
        for n in range(2):
            hydrogen_to_delete = pdb_joiner.get_H_bonded_to_atom(core_bond[0].name, core_structure,
                                                                 bond_dist=new_distance,
                                                                 banned_hydrogen=name_to_replace_core,
                                                                 chain=chain_complex)
            names_to_keep = list(core_structure.getNames())
            names_to_keep.remove(hydrogen_to_delete)
            core_structure = core_structure.select("name {}".format(" ".join(names_to_keep)))

    fragment_structure.setCoords(new_coords)
    merged_structure = bond(h_atom_names, [core_structure, fragment_structure])
    return merged_structure, name_to_replace_core, name_to_replace_fragment, new_distance # new_distance to modify the clash treshold
    # in check_collisions()


def detect_fragment_bond_type(fragment_structure, atom1, atom2, dictionary_of_distances):
    coords_atom1 = find_coords_of_atom(atom1.name, fragment_structure)
    coords_atom2 = find_coords_of_atom(atom2.name, fragment_structure)
    distance = compute_distance_between_atoms(coords_atom1, coords_atom2)
    for bond, bond_distance in dictionary_of_distances.items():
        if bond[0] == atom1.element and bond[1] == atom2.element and bond_distance+0.06 >= distance > bond_distance-0.06:
            return bond[0], bond[1], bond[2]

def convert_pdbnames_to_rdkit_idx(molecule):
   names_relations = {}
   for atom in molecule.GetAtoms():
       names_relations[atom.GetPDBResidueInfo().GetName().strip()] = atom.GetIdx()
   return names_relations

def get_bond_types_from_mol(molecule):
    bonds = {}
    for bond in molecule.GetBonds():
        bonds[(bond.GetBeginAtomIdx(), bond.GetEndAtomIdx())] = str(bond.GetBondType()).lower()
        bonds[(bond.GetEndAtomIdx(), bond.GetBeginAtomIdx())] = str(bond.GetBondType()).lower()
    return bonds

def detect_bond_type(molecule, atom1, atom2):
    names = convert_pdbnames_to_rdkit_idx(molecule)
    bonds = get_bond_types_from_mol(molecule)
    bond_names = (names[atom1.name], names[atom2.name])
    bond_type = bonds[bond_names]
    return bond_type

def correct_bonding_distance(atom_reference, atom_to_correct, reference_structure, movil_structure, bond_type="single"):
    hvy_core_coords = find_coords_of_atom(atom_reference.name, reference_structure)
    hvy_frag_coords = find_coords_of_atom(atom_to_correct.name, movil_structure)
    try:
        new_distance = atom_constants.BONDING_DISTANCES[atom_reference.element.upper(), atom_to_correct.element.upper(), bond_type]
    except KeyError:
        raise KeyError(f"Not defined bond {atom_reference.element.upper()}-{ atom_to_correct.element.upper()}. \
please define the maestro's bond distance into the file: frag_pele/frag_pele/Growing/AddingFragHelpers/atom_constants.py")
    new_coords = modify_distance_between_structures(hvy_core_coords, hvy_frag_coords, movil_structure.getCoords(),
                                                    new_distance)
    return new_coords, new_distance


def correct_hydrogen_position(hydrogen_atom, atom_to_bond_with, structure):
    coords_atom_to_bond = atom_to_bond_with.coord
    coords_H_to_move = hydrogen_atom.getCoords()
    new_distance = atom_constants.BONDING_DISTANCES["H", atom_to_bond_with.element, "single"]
    new_coords = modify_distance_between_structures(coords_atom_to_bond, coords_H_to_move, structure.getCoords(),
                                                    new_distance)
    return new_coords, new_distance


def replace_heavy_by_hydrogen(heavy_atom, structure):
    print("HEAVY ATOM TO REPLACE DETECTED!")
    index_of_atom = find_index_by_name(heavy_atom.name, structure)
    structure[index_of_atom].setElement("H")
    hydrogen_name = autoname_atoms(structure.getNames(), "H")
    structure[index_of_atom].setName(hydrogen_name)
    print("{} has been replaced for {}.".format(heavy_atom.name, structure.getNames()[index_of_atom]))
    return index_of_atom


def find_index_by_name(name, structure):
    for n, atom in enumerate(structure.getNames()):
        if atom == name:
            return n
    print("Atom not found!")


def find_coords_of_atom(atom_to_find, structure_prody):
    for atomname, coords in zip(structure_prody.getNames(), structure_prody.getCoords()):
        if atomname == atom_to_find:
            return coords


def autoname_atoms(list_of_atom_names, element):
    dictionary_of_elements = {}
    for name in list_of_atom_names:
        element_of_name = " ".join(re.split("[^a-zA-Z]*", name)).strip()
        if element_of_name in dictionary_of_elements.keys():
            dictionary_of_elements[element_of_name] += 1
        else:
            dictionary_of_elements[element_of_name] = 1
    while True:
        final_name = "{}{}".format(element, dictionary_of_elements[element])
        if final_name not in list_of_atom_names:
            return final_name
        else:
            dictionary_of_elements[element] += 1


def find_coords_of_atom(atom_to_find, structure_prody):
    try:
        for atomname, coords in zip(structure_prody.getNames(), structure_prody.getCoords()):
            if atomname == atom_to_find:
                return coords
    except AttributeError:  # If we get the attribute error probably the structure has a single atom.
        coords = structure_prody.getCoords()
        return coords


def compute_vector_between_atoms(coords_atom_1, coords_atom_2):
    vector = coords_atom_2 - coords_atom_1
    return vector  # Vector from point 1 to point 2


def compute_distance_between_atoms(coords_atom_1, coords_atom_2):
    vector = compute_vector_between_atoms(coords_atom_1, coords_atom_2)
    module = np.linalg.norm(vector)
    return module


def compute_unit_vector_between_atoms(coords_atom_1, coords_atom_2):
    vector = compute_vector_between_atoms(coords_atom_1, coords_atom_2)
    module = np.linalg.norm(vector)
    unit_vector = vector / module
    return unit_vector


def modify_distance_between_structures(coords_core, coords_fragment, coords_to_move, new_distance):
    unit_vector = compute_unit_vector_between_atoms(coords_core, coords_fragment)
    new_point = coords_core + (unit_vector*new_distance)
    vector_to_add = compute_vector_between_atoms(coords_fragment, new_point)
    new_coords = coords_to_move + vector_to_add
    return new_coords


def rotation_thought_axis(bond, theta, core_bond, list_of_atoms, fragment_bond, core_structure, fragment_structure,
                          pdb_complex, pdb_fragment, chain_complex, chain_fragment, output_path, only_grow=False):
    """
    Given a core molecule and a fragment, this function rotates the fragment atoms a certain theta angle around an axis
    (set by the bond).
    :param bond: Bio.PDB.Atom list composed by two elements: [heavy atom of the core, heavy atom of the fragment]
    :param theta: Rotation angle in rads.
    :param core_bond: Bio.PDB.Atom list with two elements: [heavy atom, hydrogen atom]. These two atoms have to be the
    ones participating in the bond of the core that we would like to use as linking point between core and fragment.
    :param list_of_atoms: list_of_atoms: Bio.PDB.Atom list with all the atoms of the fragment. These atoms have to contain
    coordinates.
    :param fragment_bond: Bio.PDB.Atom list with two elements: [hydrogen atom, heavy atom]. These two atoms have to be the
    ones participating in the bond of the fragment that we would like to use as linking point between fragment and core.
    :param core_structure: ProDy molecule that contain only the core ligand.
    :param fragment_structure: ProDy molecule that contain only the fragment ligand.
    :return: ProDy molecule with the core_structure and the fragment_structure rotated around the axis of the bond.
    """
    # Obtain the axis that we want to use as reference for the rotation
    vector = bond[1].get_vector() - bond[0].get_vector()
    # Obtain the rotation matrix for the vector (axis) and the angle (theta)
    rot_mat = bio.rotaxis(theta, vector)
    for atom in list_of_atoms:
        # Multiply the matrix of coordinates for the transpose of the rotation matrix to get the coordinates rotated
        atom.transform(rot_mat, (0, 0, 0))
        transform_coords_from_bio2prody(fragment_structure, list_of_atoms)
    rotated_structure = join_structures_to_rotate(core_bond, fragment_bond, list_of_atoms, core_structure, fragment_structure)
    return rotated_structure


def rotate_throught_bond(bond, angle, rotated_atoms, atoms_fixed):
    # Obtain the axis that we want to use as reference for the rotation
    vector = bond.getCoords()[0] - bond.getCoords()[1]
    vector = bio.Vector(vector)
    # Obtain the rotation matrix for the vector (axis) and the angle (theta)
    rot_mat = bio.rotaxis(angle, vector)
    new_coords = []
    for coords in rotated_atoms.getCoords():
        new_coord = np.dot(coords, rot_mat)
        new_coords.append(new_coord)
    rotated_atoms.setCoords(new_coords)
    structure_result = atoms_fixed + rotated_atoms
    return structure_result


def check_collision(merged_structure, bond, theta, theta_interval, core_bond, list_of_atoms, fragment_bond,
                    core_structure, fragment_structure, pdb_complex, pdb_fragment, chain_complex, chain_fragment,
                    output_path, threshold_clash=1.70, only_grow=False):
    """
    Given a structure composed by a core and a fragment, it checks that there is not collisions between the atoms of
    both. If it finds a collision, the molecule will be rotated "theta_interval" radians and the checking will be
    repeated. If it is not possible to find a conformation without atom collisions, it will print a warning.
    :param merged_structure: ProDy molecule with the core_structure and the fragment_structure concatenated.
    :param bond: Bio.PDB.Atom list composed by two elements: [heavy atom of the core, heavy atom of the fragment]
    :param theta: Initial rotation angle in rads.
    :param theta_interval: Rotation angle that will be added to theta.
    :param core_bond: Bio.PDB.Atom list with two elements: [heavy atom, hydrogen atom]. These two atoms have to be the
    ones participating in the bond of the core that we would like to use as linking point between core and fragment.
    :param list_of_atoms: list_of_atoms: Bio.PDB.Atom list with all the atoms of the fragment. These atoms have to contain
    coordinates.
    :param fragment_bond: Bio.PDB.Atom list with two elements: [hydrogen atom, heavy atom]. These two atoms have to be the
    ones participating in the bond of the fragment that we would like to use as linking point between fragment and core.
    :param core_structure: ProDy molecule that contain only the core ligand.
    :param fragment_structure: ProDy molecule that contain only the fragment ligand.
    :return: ProDy molecule with the core_structure and the fragment_structure (rotated and without intra-molecular
    clashes) around the axis of the bond.
    """
    core_resname = bond[0].get_parent().get_resname()
    frag_resname = bond[1].get_parent().get_resname()
    if core_resname is frag_resname:
        logger.critical("The resname of the core and the fragment is the same. Please, change one of both")
    print("resname {} and within {} of resname {}".format(core_resname,
                                                                                                       threshold_clash,
                                                                                                        frag_resname))
    check_possible_collision = merged_structure.select("resname {} and within {} of resname {}".format(core_resname,
                                                                                                       threshold_clash,
                                                                                                        frag_resname))

    # This list only should have the atom of the fragment that will be bonded to the core, so if not we will have to
    # solve it
    if len(check_possible_collision.getNames()) > 1 or check_possible_collision.getNames()[0] != bond[0].name:
        print("We have a collision between atoms of the fragment {} and the core {} within {} A!" 
              " Rotating the fragment to solve it...".format(frag_resname, core_resname, threshold_clash))
        theta = theta + theta_interval
        if theta >= math.pi*2:
            print("Not possible solution, decreasing the angle of rotation...")
        else:
            rotated_structure = rotation_thought_axis(bond, theta, core_bond, list_of_atoms, fragment_bond, core_structure,
                                                      fragment_structure, pdb_complex, pdb_fragment, chain_complex, chain_fragment,
                                                      output_path=output_path, only_grow=only_grow)
            #prody.writePDB("testing_{}.pdb".format(theta), rotated_structure[0])
            recall = check_collision(merged_structure=rotated_structure[0], bond=bond, theta=theta, theta_interval=theta_interval, 
                                     core_bond=core_bond, list_of_atoms=list_of_atoms, fragment_bond=fragment_bond, core_structure=core_structure,
                                     fragment_structure=fragment_structure, pdb_complex=pdb_complex, pdb_fragment=pdb_fragment, 
                                     chain_complex=chain_complex, chain_fragment=chain_fragment,
                                     output_path=output_path, threshold_clash=threshold_clash, only_grow=only_grow)
            return recall
    else:
        return merged_structure


def get_previous_bond(structure, core_atom, core_resname):
    bond_selection = structure.select("resname {} and within 1.75 of name {}".format(core_resname, core_atom))
    return bond_selection


def finishing_joining(molecule, chain):
    """
    Given a ProDy molecule this function change the Resname of the atoms to "GRW" and the Resnum to "1". Following this
    process it is possible to transform a ProDy object with more than one element with different resnums and resnames
    into a single molecule.
    :param molecule: ProDy molecule.
    :return: ProDy molecule with Resname "GRW" and Resnum "1".
    """
    molecule.setResnames("GRW")
    molecule.setResnums(1)
    molecule.setChids(chain)


def compute_centroid(molecule):
    """
    Given a ProDy molecule, the function extract the coordinates of their atoms and compute the centroid of the
    molecule.
    :param molecule: ProDy molecule object.
    :return: centroid of the molecule, tuple(X,Y,Z).
    """
    coords = molecule.getCoords()
    x = []
    y = []
    z = []
    for coord in coords:
        x.append(float(coord[0]))
        y.append(float(coord[1]))
        z.append(float(coord[2]))
    centroid = (np.mean(x),  np.mean(y), np.mean(z))
    return centroid


def move_atom_along_vector(initial_coord, final_coord, position_proportion):
    """
    Given two points (atom coordinates), this function moves the initial point a distance of "length of the vector
    formed by the two coordinates" * "position_proportion" on the vector's direction.
    :param initial_coord: initial 3D coordinates (X, Y, Z). numpy.ndarray
    :param final_coord: final 3D coordinates (X, Y, Z). numpy.ndarray
    :param position_proportion: proportion of movement that we would like to apply on the initial atom on the vector's
    direction. float(generally between 0 and 1)
    :return:
    """
    vector = final_coord - initial_coord
    new_coords = initial_coord + (position_proportion * vector)
    return new_coords


def reduce_molecule_size(molecule, residue, steps):
    """
    This function performs a reduction of the size of a given residue of a ProDy molecule object.
    :param molecule: ProDy molecule object.
    :param residue: Resname of the residue of the molecule that we want to reduce. string
    :param proportion: proportion of reduction of the size that we want to apply to the selected residue (between 0 and
    1). float
    :return: modify the coordinates of the selected residue for the result of the reduction.
    """
    proportion = 1-(1/(steps+1))
    if proportion >= 0 and proportion <= 1:
        selection = molecule.select("resname {}".format(residue))
        centroid = compute_centroid(selection)
        for atom in selection:
            atom_coords = atom.getCoords()
            new_coords = move_atom_along_vector(atom_coords, centroid, proportion)
            atom.setCoords(new_coords)
    else:
        logger.critical("Sorry, reduce_molecule_size() needs a proportion value between 0 and 1!")


def translate_to_position(initial_pos, final_pos, molecule):
    """
    This function applies a translation of a whole molecule, using the vector from the initial_pos to the final_pos.
    :param initial_pos: initial position in 3D coordinates. Generally we use the coordinates of an atom. numpy. ndarray
    :param final_pos: final position in 3D coordinates. Generally we use the coordinates of an atom. numpy. ndarray
    :param molecule: ProDy molecule object.
    :return: modify the coordinates of the molecule.
    """
    translation = initial_pos - final_pos
    coords_to_move = molecule.getCoords()
    list_of_new_coords = []
    for coords in coords_to_move:
        new_coords = coords + translation
        list_of_new_coords.append(new_coords[0])
    molecule.setCoords(list_of_new_coords)


def extract_protein_from_complex(pdb_file):
    """
    Given a pdb file containing a complex (ligand + protein) it returns only the protein.
    :param pdb_file: pdb file with a complex. string.
    :return: ProDy molecule with only the protein.
    """
    complex = prody.parsePDB(pdb_file)
    protein = complex.select("protein")
    return protein


def get_waters_or_ions_in_pdb(pdb_input):
    """
    Given a pdb file return a string with the waters contained in the file.
    :param pdb_input: pdb input file
    :return: part of the pdb that contains the waters
    """
    water_lines = []
    ion_lines = []
    with open(pdb_input) as pdb:
        for line in pdb:
            if line.startswith("HETATM"):
                if line[21] == "A" and line[17:20].split()[0] == "HOH":
                    water_lines.append(line)
                elif line[21] == "A" and line[17:20].split()[0] in LIST_OF_IONS:
                    ion_lines.append(line)
    if len(water_lines) > 0 and len(ion_lines) == 0:
        water = "".join(water + "TER\n" * (n % 3 == 2) for n, water in enumerate(water_lines))
        return water
    if len(ion_lines) > 0 and len(water_lines) == 0:
        ions = "".join(ion + "TER\n" for ion in ion_lines)
        return ions
    if len(ion_lines) > 0 and len(water_lines) > 0:
        water = "".join(water + "TER\n" * (n % 3 == 2) for n, water in enumerate(water_lines))
        ions = "".join(ion + "TER\n" for ion in ion_lines)
        return "".join(ions+water)


def get_everything_except_ligand(pdb_input, ligand_chain):
    pdb_content = []
    with open(pdb_input) as pdb:
        for line in pdb:
            if (line.startswith("ATOM") or line.startswith("HETATM") and line[21] != ligand_chain) or line.startswith("TER"):
                pdb_content.append(line)
    return "".join(pdb_content)


def check_water(pdb_input):
    """
    Given a pdb file checks if it contains water molecules.
    :param pdb_input: pdb input file
    :return: True or False
    """
    checker = False
    with open(pdb_input) as pdb:
        for line in pdb:
            if "HETATM" in line:
                if line.split()[3] == "HOH":
                    print("Your pdb file contains water molecules")
                    checker = True
                    break
    return checker


def lignames_replacer(pdb_file, original_ligname, new_ligname):
    """
    Given a PDB file it replace the name of a ligand for a new one.
    :param pdb_file: file in PDB format
    :param original_ligname: original name of the ligand
    :param new_ligname: new name of the ligand that will replace the original name
    :return:
    """
    with open(pdb_file) as pdb:
        content = pdb.readlines()
    for index, line in enumerate(content):
        if line.startswith("HETATM"):
            line = line.replace(original_ligname, new_ligname)
            content[index] = line
    pdb_modified = "".join(content)
    with open(pdb_file, "w") as overwrite_pdb:
        overwrite_pdb.write(pdb_modified)


def check_and_fix_repeated_lignames(pdb1, pdb2, ligand_chain_1="L", ligand_chain_2="L"):
    """
    It checks if two pdbs have the same ligand name or if the pdb file 1 has as ligand name "GRW" and it is replaced
    by "LIG".
    :param pdb1: pdb file 1
    :param pdb2: pdb file 2
    :return:
    """
    ligname_1 = extract_heteroatoms_pdbs(pdb1, create_file=False, ligand_chain=ligand_chain_1)
    ligname_2 = extract_heteroatoms_pdbs(pdb2, create_file=False, ligand_chain=ligand_chain_2)
    if ligname_1 == ligname_2 or ligname_1 == "GRW":
        logging.warning("REPEATED NAMES IN LIGANDS FOR THE FILES: '{}' and '{}'. {} replaced by LIG ".format(pdb1, pdb2, ligname_1))
        lignames_replacer(pdb1, ligname_1, "LIG")


def main(pdb_complex_core, pdb_fragment, pdb_atom_core_name, pdb_atom_fragment_name, steps, core_chain="L",
         fragment_chain="L", output_file_to_tmpl="growing_result.pdb", output_file_to_grow="initialization_grow.pdb",
         h_core=None, h_frag=None, rename=False, threshold_clash=1.70, output_path=None, only_grow=False):
    """
    From a core (protein + ligand core = core_chain) and fragment (fragment_chain) pdb files, given the heavy atoms
    names that we want to connect, this function add the fragment to the core structure. We will get three PDB files:
    (1) the ligand core of the complex isolated, that will be used in further steps to generate the template of the
    initial structure; (2) the ligand completed with the core and the fragment added, also prepared to generate the
    template of the final structure; (3) the pdb file that will be used to initialise PELE simulations. Here we have
    the core structure with the fragment added, but this fragment has been size-reduced in order to get small bond
    lengths between its atoms. (During PELE simulations this distance will increase linearly until it reaches the
    bond length given by the template of the final structure)
    :param pdb_complex_core: pdb file with a complex (protein + ligand) that will be used as core to perform the
    addition of the fragment. The chain of the ligand needs to be named as "L". We will also use the information of the
    protein to perform calculations of contacts with the ligand.
    :param pdb_fragment: pdb file, normally with only the ligand (please, put "L" as name of the chain that contain the
    ligand), that will be added to the core.
    :param pdb_atom_core_name: heavy atom name (string) of the ligand core where we want to add the fragment and
    form a new bond.
    :param pdb_atom_fragment_name: heavy atom name (string) of the ligand fragment where we want to perform the
    connection to form a new bond with the core.
    :param core_chain: name of the chain which contains the ligand in the pdb of the core complex. string
    :param fragment_chain: name of the chain which contains the ligand in the pdb of the fragment. string
    :param output_file_to_tmpl: name of the pdb file with the result of the connection between the core and the fragment
    (single ligand). string. The resname of the molecule will be "GRW" and the resnum "1". "growing_result.pdb" by
    default.
    :param output_file_to_grow: name of the pdb file that will be used to initialise PELE simulations. string.
    "initialization_grow.pdb" by default.
    :param h_core: if the user wants to select an specific hydrogen atom of the core to create the new bond, its name
    must be specified here.
    :param h_frag: if the user wants to select an specific hydrogen atom of the fragment to create the new bond, its name
    must be specified here.
    :param rename: if set, the names of the pdb atom names will be replaced with "G+atom_number_fragment".
    :param threshold_clash: distance that will be used to identity which atoms are doing clashes between atoms of the
    fragment and the core.
    :returns: [changing_names_dictionary, hydrogen_atoms, "{}.pdb".format(core_residue_name), output_file_to_tmpl,
    output_file_to_grow, core_original_atom, fragment_original_atom]

    """
    WORK_PATH = os.path.join(output_path, c.PRE_WORKING_DIR)
    if not os.path.exists(WORK_PATH):
        os.mkdir(WORK_PATH)
    # Check that ligand names are not repeated
    check_and_fix_repeated_lignames(pdb_complex_core, pdb_fragment, core_chain, fragment_chain)
    for pdb_file in (pdb_complex_core, pdb_fragment):
        logging.info("Checking {} ...".format(pdb_file))
        checker.check_and_fix_pdbatomnames(pdb_file)
    # Get the selected chain from the core and the fragment and convert them into ProDy molecules.
    ligand_core = complex_to_prody.pdb_parser_ligand(pdb_complex_core, core_chain)
    fragment = complex_to_prody.pdb_parser_ligand(pdb_fragment, fragment_chain)
    # We will check that the structures are protonated. We will also create a new PDB file for each one and we will get
    # the residue name of each ligand.
    core_residue_name = extract_heteroatoms_pdbs(pdb_complex_core, True, core_chain, output_folder=WORK_PATH)
    frag_residue_name = extract_heteroatoms_pdbs(pdb_fragment, True, fragment_chain, output_folder=WORK_PATH)
    # We will use the PDBs previously generated to get a list of Bio.PDB.Atoms for each structure
    bioatoms_core_and_frag = from_pdb_to_bioatomlist([os.path.join(WORK_PATH, core_residue_name),
                                                     os.path.join(WORK_PATH, frag_residue_name)])
    # Then, we will have to transform the atom names of the core and the fragment to a list object
    # (format required by functions)
    pdb_atom_names = [pdb_atom_core_name, pdb_atom_fragment_name]
    # Using the Bio.PDB.Atoms lists and this names we will get the heavy atoms that we will use later to do the bonding
    heavy_atoms = extract_heavy_atoms(pdb_atom_names, bioatoms_core_and_frag)
    # Once we have the heavy atoms, for each structure we will obtain the hydrogens bonded to each heavy atom.
    # We will need pdbs because we will use the information of the protein to select the hydrogens properly.
    hydrogen_atoms = extract_hydrogens(pdb_atom_names, bioatoms_core_and_frag, [pdb_complex_core, pdb_fragment], h_core,
                                       h_frag, core_chain, fragment_chain)
    # Create a list with the atoms that form a bond in core and fragment.
    core_bond = [heavy_atoms[0], hydrogen_atoms[0]]
    fragment_bond = [hydrogen_atoms[1], heavy_atoms[1]]  # This has to be in inverted order to do correctly the superimposition
    logger.info("Performing a superimposition of bond {} of the fragment on bond {} of the core..."
                .format(fragment_bond, core_bond))
    # Using the previous information we will superimpose the whole fragment on the bond of the core in order to place
    # the fragment in the correct position, deleting the H.
    merged_structure, core_original_atom, fragment_original_atom, new_dist = join_structures(core_bond, fragment_bond,
                                                                                             ligand_core, fragment,
                                                                                             pdb_complex_core, pdb_fragment,
                                                                                             core_chain, fragment_chain,
                                                                                             output_path=WORK_PATH, only_grow=only_grow)
    prody.writePDB("merged.pdb", merged_structure[0])
    if not only_grow:
        # It is possible to create intramolecular clashes after placing the fragment on the bond of the core, so we will
        # check if this is happening, and if it is, we will perform rotations of 10º until avoid the clash.
        # check_results = check_collision(merged_structure[0], heavy_atoms, 0, math.pi/18, core_bond,
        #                            bioatoms_core_and_frag[1], fragment_bond, ligand_core, fragment)
        # check_collision(merged_structure, bond, theta, theta_interval, core_bond, list_of_atoms, fragment_bond,
        #            core_structure, fragment_structure)
        check_results = check_collision(merged_structure=merged_structure[0], bond=heavy_atoms, theta=0,
                                        theta_interval=math.pi/18, core_bond=core_bond, list_of_atoms=bioatoms_core_and_frag[1],
                                        fragment_bond=fragment_bond, core_structure=ligand_core, fragment_structure=fragment,
                                        pdb_complex=pdb_complex_core, pdb_fragment=pdb_fragment, chain_complex=core_chain,
                                        chain_fragment=fragment_chain, output_path=WORK_PATH, threshold_clash=new_dist+0.01,
                                        only_grow=only_grow)
        # If we do not find a solution in the previous step, we will repeat the rotations applying only increments of 1º
        if not check_results:
            check_results = check_collision(merged_structure=merged_structure[0], bond=heavy_atoms, theta=0,
                                            theta_interval=math.pi/180, core_bond=core_bond, list_of_atoms=bioatoms_core_and_frag[1],
                                            fragment_bond=fragment_bond, core_structure=ligand_core, fragment_structure=fragment,
                                            pdb_complex=pdb_complex_core, pdb_fragment=pdb_fragment, chain_complex=core_chain,
                                            chain_fragment=fragment_chain, output_path=WORK_PATH,
                                            threshold_clash=new_dist+0.01, only_grow=only_grow)
        # Now, we want to extract this structure in a PDB to create the template file after the growing. We will do a copy
        # of the structure because then we will need to resize the fragment part, so be need to keep it as two different
        # residues.
        try:
            structure_to_template = check_results.copy()
        except AttributeError:
            raise AttributeError("Frag cannot superimpose the fragment onto the core's hydrogen.  \
In order to create space for the fragment \
manually rotate the hydrogen bond of the core where the fragment will be attached to.   \
We are currently working to fix this automatically")

        # Once we have all the atom names unique, we will rename the resname and the resnum of both, core and fragment, to
        # GRW and 1. Doing this, the molecule composed by two parts will be transformed into a single one.
        changing_names = pdb_joiner.extract_and_change_atomnames(structure_to_template, fragment.getResnames()[0],
                                                                 core_residue_name, rename=rename)
        molecule_names_changed, changing_names_dictionary = changing_names

        # Check if there is still overlapping names
        if pdb_joiner.check_overlapping_names(molecule_names_changed):
            logger.critical("{} is repeated in the fragment and the core. Please, change this atom name of the core by"
                            " another one.".format(pdb_joiner.check_overlapping_names(molecule_names_changed)))
        logger.info("The following names of the fragment have been changed:")
        for transformation in changing_names_dictionary:
            logger.info("{} --> {}".format(transformation, changing_names_dictionary[transformation]))
        finishing_joining(molecule_names_changed, core_chain)
        # Extract a PDB file to do the templates
        prody.writePDB(os.path.join(WORK_PATH, output_file_to_tmpl), molecule_names_changed)
        logger.info("The result of core + fragment has been saved in '{}'. This will be used to create the template file."
                    .format(os.path.join(WORK_PATH, output_file_to_tmpl)))
        # Now, we will use the original molecule to do the resizing of the fragment.
        reduce_molecule_size(check_results, frag_residue_name, steps)
        point_reference = check_results.select("name {} and resname {}".format(pdb_atom_fragment_name, frag_residue_name))
        fragment_segment = check_results.select("resname {}".format(frag_residue_name))
        translate_to_position(hydrogen_atoms[0].get_coord(), point_reference.getCoords(), fragment_segment)

        # Repeat all the preparation process to finish the writing of the molecule.
        changing_names = pdb_joiner.extract_and_change_atomnames(check_results, fragment.getResnames()[0], core_residue_name, rename=rename)
        molecule_names_changed, changing_names_dictionary = changing_names
        finishing_joining(molecule_names_changed, core_chain)
        logger.info("The result of core + fragment(small) has been saved in '{}'. This will be used to initialise the growing."
                    .format(os.path.join(WORK_PATH, output_file_to_grow)))
        # Add the protein to the ligand
        output_ligand_grown_path = os.path.join(WORK_PATH, "ligand_grown.pdb")
        prody.writePDB(output_ligand_grown_path, molecule_names_changed)

        with open(output_ligand_grown_path) as lig:
            content_lig = lig.readlines()
            content_lig = content_lig[1:]
            content_lig = "".join(content_lig)

        # Join all parts of the PDB
        output_file = []
        chain_not_lig = get_everything_except_ligand(pdb_complex_core, core_chain)
        output_file.append(chain_not_lig)
        output_file.append("{}TER".format(content_lig))
        out_joined = "".join(output_file)
        with open(os.path.join(WORK_PATH, output_file_to_grow), "w") as output: # Save the file in the pregrow folder
            output.write(out_joined)
        # Make a copy of output files in the main directory
        shutil.copy(os.path.join(WORK_PATH, output_file_to_grow), ".")  # We assume that the user will be running FrAG in PELE's main folder...
        # In further steps we will probably need to recover the names of the atoms for the fragment, so for this reason we
        # are returning this dictionary in the function.
        with open(os.path.join(WORK_PATH, "changingatoms.dict"), "wb") as pkl:
            pickle.dump(changing_names_dictionary, pkl)
    else:
        with open( os.path.join(WORK_PATH, "changingatoms.dict"), "rb") as pkl:
            changing_names_dictionary = pickle.load(pkl)


    return changing_names_dictionary, hydrogen_atoms, "{}.pdb".format(core_residue_name), \
           os.path.join(WORK_PATH, output_file_to_tmpl), \
           os.path.join(WORK_PATH, output_file_to_grow), \
           core_original_atom, fragment_original_atom





import logging

# Getting the name of the module for the log system
logger = logging.getLogger(__name__)

# Definition of reggex patterns
HEADER_OPLS2005 = "* LIGAND DATABASE FILE (OPLS2005)\n*\n"
PATTERN_OPLS2005_RESX_HEADER = "{:5} {:6d} {:6d} {:7d} {:7d} {:8d} \n"
PATTERN_OPLS2005_RESX_LINE = "{:5d} {:5d} {:1}   {:4} {:4} {:5d} {: >11.5f} {: >11.5f} {: >11.5f}\n"
PATTERN_OPLS2005_NBON = "{:5d} {: >8.4f} {: >8.4f} {: >10.6f} {: >8.4f} {: >8.4f} {: >13.9f} {: >13.9f}\n"
PATTERN_OPLS2005_BOND = "{:5d} {:5d} {:>9.3f} {:>6.3f}\n"
PATTERN_OPLS2005_THETA = "{:5d} {:5d} {:5d} {:>11.5f} {: >11.5f}\n"
PATTERN_OPLS2005_PHI = "{:5d} {:5d} {: 5d} {:5d} {:>9.5f} {: >4.1f} {: >3.1f}\n"


class Atom:
    """A class which contains all the information and properties of an Atom, and several methods to build templates from
    this data (currently only in OPLS2005)."""
    def __init__(self, atom_id, parent_id, location, atom_type, pdb_atom_name, unknown, x_zmatrix=0, y_zmatrix=0,
                 z_zmatrix=0, sigma=0, epsilon=0, charge=0, radnpSGB=0, radnpType=0, sgbnpGamma=0, sgbnpType=0,
                 is_linker=False, is_fragment=False):
        """
        :param atom_id: ID of the atom in the template.
        :type atom_id: int
        :param parent_id: ID ot the parent atom.
        :type parent_id: int
        :param location: Location of the atom. M for backbone, S for side chain; ligand links and non-ligand links use
        these classification for some checks. However, protein links define its backbone atoms independently of this
        flag; protein residue templates have M only for N, C and CA, under whatever name is actually used for those
        atoms.
        :type location: str
        :param atom_type: Atom type. For example: CA (carbon aromatic), CT (carbon terminal), etc.
        :type atom_type: str
        :param pdb_atom_name: PDB atom name.
        :type pdb_atom_name: str
        :param unknown: Nobody knows what is this...
        :type unknown: int
        :param x_zmatrix: Coord X of the Z-matrix.
        :type x_zmatrix: float
        :param y_zmatrix: Coord Y of the Z-matrix.
        :type y_zmatrix: float
        :param z_zmatrix: Coord Z of the Z-matrix.
        :type z_zmatrix: float
        :param sigma: sigma value, used to compute Van Der Waals terms. Units in Armstrong.
        :type sigma: float
        :param epsilon: epsilon value, used to compute Van Der Waals terms. Units in kcal/mol.
        :type epsilon: float
        :param charge: charge value, used to compute electrostatic potentials. Units are elementary charge.
        :type charge: float
        :param radnpSGB: radii of non polar SGB. Atomic radii used to calculate the surface of the molecules when
        obtaining SGB Born radii.
        :type radnpSGB: float
        :param radnpType: radii of non polar Type. Atomic radii used to calculate SASA in the non-polar term of the SGB
        and VDGBNP models
        :type radnpType: float
        :param sgbnpGamma: SGB non polar Gamma. Gamma parameter of the nonpolar model.
        :type sgbnpGamma: float
        :param sgbnpType: SGB non polar Type. Alpha parameter for the nonpolar model
        :param is_linker: Flag set when the atom is linking the fragment and the core.
        :type is_linker: bool
        :param is_fragment: Flag set when the atom is of the fragment.
        :type is_fragment: bool
        """
        self.atom_id = int(atom_id)
        self.parent_id = int(parent_id)
        self.location = str(location)
        self.atom_type = str(atom_type)
        self.pdb_atom_name = str(pdb_atom_name)
        self.unknown = int(unknown)
        self.x_zmatrix = float(x_zmatrix)
        self.y_zmatrix = float(y_zmatrix)
        self.z_zmatrix = float(z_zmatrix)
        self.sigma = float(sigma)
        self.epsilon = float(epsilon)
        self.charge = float(charge)
        self.radnpSGB = float(radnpSGB)
        self.radnpType = float(radnpType)
        self.sgbnpGamma = float(sgbnpGamma)
        self.sgbnpType = float(sgbnpType)
        self.bonds = []
        self.thetas = []
        self.phis = []
        self.iphis = []
        self.is_fragment = bool(is_fragment)
        self.is_linker = bool(is_linker)

    def write_resx(self):
        return PATTERN_OPLS2005_RESX_LINE.format(self.atom_id, self.parent_id, self.location, self.atom_type.strip(),
                                                 self.pdb_atom_name, self.unknown, self.x_zmatrix,
                                                 self.y_zmatrix, self.z_zmatrix)

    def write_nbon(self):
        return PATTERN_OPLS2005_NBON.format(self.atom_id, self.sigma, self.epsilon, self.charge, self.radnpSGB,
                                            self.radnpType, self.sgbnpGamma, self.sgbnpType)


class Bond:
    def __init__(self, atom1, atom2, spring, eq_dist, is_fragment=False, is_linker=False):
        self.atom1 = int(atom1)
        self.atom2 = int(atom2)
        self.spring = float(spring)
        self.eq_dist = float(eq_dist)
        self.is_fragment = bool(is_fragment)
        self.is_linker = bool(is_linker)

    def write_bond(self):
        return PATTERN_OPLS2005_BOND.format(self.atom1, self.atom2, self.spring, self.eq_dist)


class Theta:
    def __init__(self, atom1, atom2, atom3, spring, eq_angle, is_fragment=False):
        self.atom1 = int(atom1)
        self.atom2 = int(atom2)
        self.atom3 = int(atom3)
        self.spring = float(spring)
        self.eq_angle = float(eq_angle)
        self.is_fragment = bool(is_fragment)

    def write_theta(self):
        return PATTERN_OPLS2005_THETA.format(self.atom1, self.atom2, self.atom3, self.spring, self.eq_angle)


class Phi:
    def __init__(self, atom1, atom2, atom3, atom4, constant, prefactor, nterm, improper, is_fragment=False):
        self.atom1 = int(atom1)
        self.atom2 = int(atom2)
        self.atom3 = int(atom3)
        self.atom4 = int(atom4)
        self.constant = float(constant)
        self.prefactor = float(prefactor)
        self.nterm = float(nterm)
        self.improper = bool(improper)
        self.is_fragment = bool(is_fragment)

    def write_phi(self):
        if not self.improper:
            return PATTERN_OPLS2005_PHI.format(self.atom1, self.atom2, self.atom3, self.atom4, self.constant,
                                               self.prefactor, self.nterm)

    def write_iphi(self):
        if self.improper:
            return PATTERN_OPLS2005_PHI.format(self.atom1, self.atom2, self.atom3, self.atom4, self.constant,
                                               self.prefactor, self.nterm)


class TemplateOPLS2005:
    def __init__(self, path_to_template):
        self.path_to_template = path_to_template
        self.template_name = ""
        self.num_nbon_params = 0
        self.num_bond_params = 0
        self.num_angle_params = 0
        self.num_dihedr_params = 0
        self.num_nonnull = 0
        self.list_of_atoms = {}
        self.list_of_bonds = {}
        self.list_of_thetas = {}
        self.list_of_phis = []
        self.list_of_iphis = []
        self.unique_atoms = []
        self.read_template()

    def read_template(self):
        template = file_to_list_of_lines(self.path_to_template)
        for line in template[2:3]:
            self.template_name = get_string_from_line(line=line, index_initial=0, index_final=5)
            self.num_nbon_params = int(get_string_from_line(line=line, index_initial=6, index_final=11))
            self.num_bond_params = int(get_string_from_line(line=line, index_initial=13, index_final=17))
            self.num_angle_params = int(get_string_from_line(line=line, index_initial=18, index_final=24))
            self.num_dihedr_params = int(get_string_from_line(line=line, index_initial=25, index_final=31))
            self.num_nonnull = int(get_string_from_line(line=line, index_initial=32, index_final=39))
        for line in template[3:]:
            if line.startswith("NBON"):
                index = template.index(line)
                break
            try:
                atom_id = get_string_from_line(line=line, index_initial=0, index_final=6)
                parent_id = get_string_from_line(line=line, index_initial=6, index_final=11)
                location = get_string_from_line(line=line, index_initial=12, index_final=13)
                atom_type = get_string_from_line(line=line, index_initial=15, index_final=20)
                pdb_atom_name = get_string_from_line(line=line, index_initial=21, index_final=25)
                unknown = get_string_from_line(line=line, index_initial=26, index_final=31)
                x_zmatrix = get_string_from_line(line=line, index_initial=32, index_final=43)
                y_zmatrix = get_string_from_line(line=line, index_initial=44, index_final=55)
                z_zmatrix = get_string_from_line(line=line, index_initial=56, index_final=67)
                atom = Atom(atom_id=atom_id, parent_id=parent_id, location=location, atom_type=atom_type,
                            pdb_atom_name=pdb_atom_name, unknown=unknown, x_zmatrix=x_zmatrix, y_zmatrix=y_zmatrix,
                            z_zmatrix=z_zmatrix)
                self.list_of_atoms.setdefault(atom.atom_id, atom)
                if pdb_atom_name not in self.unique_atoms:
                    self.unique_atoms.append(pdb_atom_name)
                else:
                    raise ValueError("ERROR: PDB ATOM NAME {} ALREADY EXISTS in the template {}!".format(pdb_atom_name,
                                                                                              self.path_to_template))
            except ValueError:
                raise ValueError(
                     "Unexpected type in line {} of {}\n{}".format(template.index(line), self.path_to_template, line))

        for line in template[index + 1:]:
            if line.startswith("BOND"):
                index = template.index(line)
                break
            try:
                id = int(get_string_from_line(line=line, index_initial=0, index_final=6))
                self.list_of_atoms[id].sigma = float(get_string_from_line(line=line, index_initial=7, index_final=14))
                self.list_of_atoms[id].epsilon = float(get_string_from_line(line=line, index_initial=15, index_final=23))
                self.list_of_atoms[id].charge = float(get_string_from_line(line=line, index_initial=24, index_final=34))
                self.list_of_atoms[id].radnpSGB = float(get_string_from_line(line=line, index_initial=35, index_final=43))
                self.list_of_atoms[id].radnpType = float(get_string_from_line(line=line, index_initial=44, index_final=52))
                self.list_of_atoms[id].sgbnpGamma = float(get_string_from_line(line=line, index_initial=53,
                                                                               index_final=66))
                self.list_of_atoms[id].sgbnpType = float(get_string_from_line(line=line, index_initial=67, index_final=80))
            except ValueError:
                raise ValueError(
                    "Unexpected type in line {} of {}\n{}".format(template.index(line), self.path_to_template, line))

        for line in template[index + 1:]:
            if line.startswith("THET"):
                index = template.index(line)
                break
            try:
                id_atom1 = int(get_string_from_line(line=line, index_initial=0, index_final=6))
                id_atom2 = int(get_string_from_line(line=line, index_initial=6, index_final=12))
                spring = get_string_from_line(line=line, index_initial=13, index_final=21)
                eq_dist = get_string_from_line(line=line, index_initial=23, index_final=29)
                # Create bond instance
                bond = Bond(atom1=id_atom1, atom2=id_atom2, spring=spring, eq_dist=eq_dist)
                self.list_of_bonds.setdefault((id_atom1, id_atom2), bond)
                # Set which atom is bonded with
                self.list_of_atoms[id_atom1].bonds.append(bond)

            except ValueError:
                raise ValueError(
                    "Unexpected type in line {} of {}\n{}".format(template.index(line), self.path_to_template, line))
        for line in template[index + 1:]:
            if line.startswith("PHI"):
                index = template.index(line)
                break
            try:
                id_atom1 = int(get_string_from_line(line=line, index_initial=0, index_final=6))
                id_atom2 = int(get_string_from_line(line=line, index_initial=6, index_final=12))
                id_atom3 = int(get_string_from_line(line=line, index_initial=13, index_final=18))
                spring = get_string_from_line(line=line, index_initial=19, index_final=29)
                eq_angle = get_string_from_line(line=line, index_initial=31, index_final=40)
                # Create bond instance
                theta = Theta(atom1=id_atom1, atom2=id_atom2, atom3=id_atom3, spring=spring, eq_angle=eq_angle)
                self.list_of_thetas.setdefault((id_atom1, id_atom2, id_atom3), theta)
                self.list_of_atoms[id_atom1].thetas.append(theta)

            except ValueError:
                raise ValueError(
                    "Unexpected type in line {} of {}\n{}".format(template.index(line), self.path_to_template,
                                                                  line))
        for line in template[index + 1:]:
            if line.startswith("IPHI"):
                index = template.index(line)
                break
            try:
                id_atom1 = int(get_string_from_line(line=line, index_initial=0, index_final=5))
                id_atom2 = int(get_string_from_line(line=line, index_initial=6, index_final=11))
                id_atom3 = int(get_string_from_line(line=line, index_initial=12, index_final=17))
                id_atom4 = int(get_string_from_line(line=line, index_initial=18, index_final=23))
                constant = get_string_from_line(line=line, index_initial=25, index_final=32)
                preafactor = get_string_from_line(line=line, index_initial=33, index_final=38)
                nterm = get_string_from_line(line=line, index_initial=39, index_final=42)
                # Create bond instance
                phi = Phi(atom1=id_atom1, atom2=id_atom2, atom3=id_atom3, atom4=id_atom4, constant=constant,
                          prefactor=preafactor, nterm=nterm, improper=False)
                self.list_of_phis.append(phi)
                self.list_of_atoms[id_atom1].phis.append(phi)

            except ValueError:
                raise ValueError(
                    "Unexpected type in line {} of {}\n{}".format(template.index(line), self.path_to_template,
                                                                  line))
        for line in template[index + 1:]:
            if line.startswith("END"):
                break
            try:
                id_atom1 = int(get_string_from_line(line=line, index_initial=0, index_final=6))
                id_atom2 = int(get_string_from_line(line=line, index_initial=7, index_final=12))
                id_atom3 = int(get_string_from_line(line=line, index_initial=13, index_final=18))
                id_atom4 = int(get_string_from_line(line=line, index_initial=19, index_final=24))
                constant = get_string_from_line(line=line, index_initial=26, index_final=34)
                preafactor = get_string_from_line(line=line, index_initial=34, index_final=39)
                nterm = get_string_from_line(line=line, index_initial=40, index_final=43)
                # Create bond instance
                phi = Phi(atom1=id_atom1, atom2=id_atom2, atom3=id_atom3, atom4=id_atom4, constant=constant,
                          prefactor=preafactor, nterm=nterm, improper=True)
                self.list_of_iphis.append(phi)

            except ValueError:
                raise ValueError(
                    "Unexpected type in line {} of {}\n{}".format(template.index(line), self.path_to_template,
                                                                  line))

    def write_header(self):
        return HEADER_OPLS2005+PATTERN_OPLS2005_RESX_HEADER.format(self.template_name, self.num_nbon_params,
                                                                   self.num_bond_params, self.num_angle_params,
                                                                   self.num_dihedr_params, self.num_nonnull)

    def write_xres(self):
        content = []
        for n in range(1, len(self.list_of_atoms)+1):
            line = self.list_of_atoms[n].write_resx()
            content.append(line)
        return "".join(content)

    def write_nbon(self):
        content = []
        for n in range(1, len(self.list_of_atoms)+1):
            line = self.list_of_atoms[n].write_nbon()
            content.append(line)
        return "".join(content)

    def write_bond(self):
        content = []
        for key in self.list_of_bonds.keys():
            line = self.list_of_bonds[key].write_bond()
            content.append(line)
        return "".join(content)

    def write_theta(self):
        content = []
        for key in self.list_of_thetas.keys():
            line = self.list_of_thetas[key].write_theta()
            content.append(line)
        return "".join(content)

    def write_phis(self):
        content = []
        for phi in self.list_of_phis:
            line = phi.write_phi()
            content.append(line)
        return "".join(content)

    def write_iphis(self):
        content = []
        for phi in self.list_of_iphis:
            line = phi.write_iphi()
            content.append(line)
        return "".join(content)

    def write_template(self):
        header = self.write_header()
        xres = self.write_xres()
        nbon_header = "NBON\n"
        nbon = self.write_nbon()
        bond_header = "BOND\n"
        bond = self.write_bond()
        theta_header = "THET\n"
        theta = self.write_theta()
        phis_header = "PHI\n"
        phis = self.write_phis()
        iphis_header = "IPHI\n"
        iphis = self.write_iphis()
        ending = "END"

        return header+xres+nbon_header+nbon+bond_header+bond+theta_header+theta+phis_header+phis+iphis_header+iphis+ending

    def write_template_to_file(self, template_new_name=None):
        if not template_new_name:
            name = self.template_name.lower()+"z"
        else:
            name = template_new_name
        with open(name, "w") as template:
            template.write(self.write_template())

    def get_list_of_fragment_atoms(self):
        atoms = []
        for key, atom in self.list_of_atoms.items():
            if atom.is_fragment:
                atoms.append((key, atom))
        return atoms
    
    def get_list_of_core_atoms(self):
        atoms = []
        for key, atom in self.list_of_atoms.items():
            if not atom.is_fragment:
                atoms.append((key, atom))
        return atoms
 
    def get_list_of_fragment_bonds(self):
        bonds = []
        for key, bond in self.list_of_bonds.items():
            if bond.is_fragment:
                bonds.append((key, bond))
        return bonds

    def get_list_of_fragment_thetas(self):
        thetas = []
        for key, theta in self.list_of_thetas.items():
            if theta.is_fragment:
                thetas.append((key, theta))
        return thetas

    def get_list_of_fragment_phis(self):
        phis = []
        for phi in self.list_of_phis:
            if phi.is_fragment:
                phis.append(phi)
        return phis

    def get_list_of_fragment_iphis(self):
        iphis = []
        for iphi in self.list_of_iphis:
            if iphi.is_fragment:
                iphis.append(iphi)
        return iphis


class ReduceProperty:
    def __init__(self, template, lambda_to_reduce, template_core=None, atom_to_replace=None):
        self.template = template
        self.lambda_to_reduce = lambda_to_reduce
        self.template_core = template_core
        self.atom_to_replace = atom_to_replace

    def reduce_epsilons(self, function):
        atoms = self.template.get_list_of_fragment_atoms()
        for key, atom in atoms:
            epsilon = atom.epsilon
            result = function(epsilon)
            self.template.list_of_atoms[key].epsilon = result

    def reduce_sigmas(self, function):
        atoms = self.template.get_list_of_fragment_atoms()
        for key, atom in atoms:
            sigma = atom.sigma
            result = function(sigma)
            self.template.list_of_atoms[key].sigma = result

    def reduce_charges(self, function):
        atoms = self.template.get_list_of_fragment_atoms()
        for key, atom in atoms:
            charge = atom.charge
            result = function(charge)
            self.template.list_of_atoms[key].charge = result
    
    def reduce_sgbnpGamma(self, function):
        atoms = self.template.get_list_of_fragment_atoms()
        for key, atom in atoms:
            sgbnpGamma = atom.sgbnpGamma
            result = function(sgbnpGamma)
            self.template.list_of_atoms[key].sgbnpGamma = result

    def reduce_sgbnpType(self, function):
        atoms = self.template.get_list_of_fragment_atoms()
        for key, atom in atoms:
            sgbnpType = atom.sgbnpType
            result = function(sgbnpType)
            self.template.list_of_atoms[key].sgbnpType = result

    def reduce_radnpSGB(self, function):
        atoms = self.template.get_list_of_fragment_atoms()
        for key, atom in atoms:
            radnpSGB = atom.radnpSGB
            result = function(radnpSGB)
            self.template.list_of_atoms[key].radnpSGB = result

    def reduce_radnpType(self, function):
        atoms = self.template.get_list_of_fragment_atoms()
        for key, atom in atoms:
            radnpType = atom.radnpType
            result = function(radnpType)
            self.template.list_of_atoms[key].radnpType = result

    def reduce_nbon_params(self, function):
        atoms = self.template.get_list_of_fragment_atoms()
        for key, atom in atoms:
            result = function(atom.epsilon)
            self.template.list_of_atoms[key].epsilon = result
            result = function(atom.sigma)
            self.template.list_of_atoms[key].sigma = result
            result = function(atom.charge)
            self.template.list_of_atoms[key].charge = result
            result = function(atom.sgbnpGamma)
            self.template.list_of_atoms[key].sgbnpGamma = result
            result = function(atom.sgbnpType)
            self.template.list_of_atoms[key].sgbnpType = result
            result = function(atom.radnpSGB)
            self.template.list_of_atoms[key].radnpSGB = result
            result = function(atom.radnpType)
            self.template.list_of_atoms[key].radnpType = result

    def reduce_bond_eq_dist(self, function):
        bonds = self.template.get_list_of_fragment_bonds()
        for key, bond in bonds:
            eq_dist = bond.eq_dist
            result = function(eq_dist)
            self.template.list_of_bonds[key].eq_dist = result

    def modify_core_epsilons(self, function):
        atoms_grw = self.template.get_list_of_core_atoms()
        atoms_ini = self.template_core.get_list_of_core_atoms()
        for atom_grow in atoms_grw:
            index_g, atom_g = atom_grow
            for atom_core in atoms_ini:
               index_c, atom_c = atom_core
               if atom_g.pdb_atom_name == atom_c.pdb_atom_name:
                   result = function(atom_c.epsilon, atom_g.epsilon)
                   self.template.list_of_atoms[index_g].epsilon = result

    def modify_core_sigmas(self, function):
        atoms_grw = self.template.get_list_of_core_atoms()
        atoms_ini = self.template_core.get_list_of_core_atoms()
        for atom_grow in atoms_grw:
            index_g, atom_g = atom_grow
            for atom_core in atoms_ini:
               index_c, atom_c = atom_core
               if atom_g.pdb_atom_name == atom_c.pdb_atom_name:
                   result = function(atom_c.sigma, atom_g.sigma)
                   self.template.list_of_atoms[index_g].sigma = result
    
    def modify_core_charges(self, function):
        atoms_grw = self.template.get_list_of_core_atoms()
        atoms_ini = self.template_core.get_list_of_core_atoms()
        for atom_grow in atoms_grw:
            index_g, atom_g = atom_grow
            for atom_core in atoms_ini:
               index_c, atom_c = atom_core
               if atom_g.pdb_atom_name == atom_c.pdb_atom_name:
                   result = function(atom_c.charge, atom_g.charge)
                   self.template.list_of_atoms[index_g].charge = result

    def modify_core_nbond_params(self, function):
        atoms_grw = self.template.get_list_of_core_atoms()
        atoms_ini = self.template_core.get_list_of_core_atoms()
        for atom_grow in atoms_grw:
            index_g, atom_g = atom_grow
            for atom_core in atoms_ini:
               index_c, atom_c = atom_core
               if atom_g.pdb_atom_name == atom_c.pdb_atom_name and not atom_g.is_fragment:
                   result = function(atom_c.epsilon, atom_g.epsilon)
                   self.template.list_of_atoms[index_g].epsilon = result
                   result = function(atom_c.sigma, atom_g.sigma)
                   self.template.list_of_atoms[index_g].sigma = result
                   result = function(atom_c.charge, atom_g.charge)
                   self.template.list_of_atoms[index_g].charge = result
                   result = function(atom_c.sgbnpGamma, atom_g.sgbnpGamma)
                   self.template.list_of_atoms[index_g].sgbnpGamma = result
                   result = function(atom_c.sgbnpType, atom_g.sgbnpType)
                   self.template.list_of_atoms[index_g].sgbnpType = result
                   result = function(atom_c.radnpSGB, atom_g.radnpSGB)
                   self.template.list_of_atoms[index_g].radnpSGB = result
                   result = function(atom_c.radnpType, atom_g.radnpType)
                   self.template.list_of_atoms[index_g].radnpType = result
                   

    def modify_core_bond_eq_dist(self, function):
        bonds_grw = self.template
        bonds_ini = self.template_core
        for bond_grw in bonds_grw.list_of_bonds.items():
            key_grw, bond_g = bond_grw
            # Get PDB atom names of the bonding atoms
            atoms_g = [bonds_grw.list_of_atoms[bond_g.atom1].pdb_atom_name, 
                       bonds_grw.list_of_atoms[bond_g.atom2].pdb_atom_name]
            for bond_ini in bonds_ini.list_of_bonds.items():
                key_ini, bond_i = bond_ini
                atoms_i = [bonds_ini.list_of_atoms[bond_i.atom1].pdb_atom_name, 
                       bonds_ini.list_of_atoms[bond_i.atom2].pdb_atom_name]
                if sorted(atoms_i) == sorted(atoms_g) and not bond_g.is_linker:
                    result = function(bond_i.eq_dist, bond_g.eq_dist)
                    bond_g.eq_dist = result
    
    def modify_core_theta(self, function):
        grw = self.template
        ini = self.template_core
        for theta_grw in grw.list_of_thetas.items():
            key_grw, theta_g = theta_grw
            # Get PDB atom names of the atoms forming angle
            atoms_g = [grw.list_of_atoms[theta_g.atom1].pdb_atom_name,
                       grw.list_of_atoms[theta_g.atom2].pdb_atom_name,
                       grw.list_of_atoms[theta_g.atom3].pdb_atom_name]
            for theta_ini in ini.list_of_thetas.items():
                key_ini, theta_i = theta_ini
                atoms_i = [ini.list_of_atoms[theta_i.atom1].pdb_atom_name,
                           ini.list_of_atoms[theta_i.atom2].pdb_atom_name,
                           ini.list_of_atoms[theta_i.atom3].pdb_atom_name]
                if sorted(atoms_i) == sorted(atoms_g):
                    result = function(theta_i.spring, theta_g.spring)
                    theta_g.spring = result
                    result = function(theta_i.eq_angle, theta_g.eq_angle)
                    theta_g.eq_angle = result

    def modify_core_phis(self, function):
        grw = self.template
        ini = self.template_core
        for phi_g in grw.list_of_phis: # This is a list, not a dict
            # Get PDB atom names of the atoms forming angles
            atoms_g = [grw.list_of_atoms[phi_g.atom1].pdb_atom_name,
                       grw.list_of_atoms[phi_g.atom2].pdb_atom_name,
                       grw.list_of_atoms[phi_g.atom3].pdb_atom_name,
                       grw.list_of_atoms[phi_g.atom4].pdb_atom_name]
            for phi_i in ini.list_of_phis:
                atoms_i = [ini.list_of_atoms[phi_i.atom1].pdb_atom_name,
                           ini.list_of_atoms[phi_i.atom2].pdb_atom_name,
                           ini.list_of_atoms[phi_i.atom3].pdb_atom_name,
                           ini.list_of_atoms[phi_i.atom4].pdb_atom_name]
                if sorted(atoms_i) == sorted(atoms_g) and phi_i.nterm == phi_g.nterm:
                    result = function(phi_i.constant, phi_g.constant)
                    phi_g.constant = result

    def modify_core_sgbnpGamma(self, function):
        atoms_grw = self.template.get_list_of_core_atoms()
        atoms_ini = self.template_core.get_list_of_core_atoms()
        for atom_grow in atoms_grw:
            index_g, atom_g = atom_grow
            for atom_core in atoms_ini:
               index_c, atom_c = atom_core
               if atom_g.pdb_atom_name == atom_c.pdb_atom_name:
                   result = function(atom_c.sgbnpGamma, atom_g.sgbnpGamma)
                   self.template.list_of_atoms[index_g].sgbnpGamma = result   

    def modify_core_sgbnpType(self, function):
        atoms_grw = self.template.get_list_of_core_atoms()
        atoms_ini = self.template_core.get_list_of_core_atoms()
        for atom_grow in atoms_grw:
            index_g, atom_g = atom_grow
            for atom_core in atoms_ini:
               index_c, atom_c = atom_core
               if atom_g.pdb_atom_name == atom_c.pdb_atom_name:
                   result = function(atom_c.sgbnpType, atom_g.sgbnpType)
                   self.template.list_of_atoms[index_g].sgbnpType = result

    def modify_core_radnpSGB(self, function):
        atoms_grw = self.template.get_list_of_core_atoms()
        atoms_ini = self.template_core.get_list_of_core_atoms()
        for atom_grow in atoms_grw:
            index_g, atom_g = atom_grow
            for atom_core in atoms_ini:
               index_c, atom_c = atom_core
               if atom_g.pdb_atom_name == atom_c.pdb_atom_name:
                   result = function(atom_c.radnpSGB, atom_g.radnpSGB)
                   self.template.list_of_atoms[index_g].radnpSGB = result

    def modify_core_radnpType(self, function):
        atoms_grw = self.template.get_list_of_core_atoms()
        atoms_ini = self.template_core.get_list_of_core_atoms()
        for atom_grow in atoms_grw:
            index_g, atom_g = atom_grow
            for atom_core in atoms_ini:
               index_c, atom_c = atom_core
               if atom_g.pdb_atom_name == atom_c.pdb_atom_name:
                   result = function(atom_c.radnpType, atom_g.radnpType)
                   self.template.list_of_atoms[index_g].radnpType = result

    def modify_linker_bond_eq_dist(self, function):
        grw = self.template
        ini = self.template_core
        atm_to_repl = self.atom_to_replace
        for bond_grw in grw.list_of_bonds.items():
            key_grw, bond_g = bond_grw
            atoms_g = [grw.list_of_atoms[bond_g.atom1].pdb_atom_name,
                       grw.list_of_atoms[bond_g.atom2].pdb_atom_name]
            # Check which one is linker
            if bond_g.is_linker:
                 # Search the atoms in the initial template
                for bond_ini in ini.list_of_bonds.items():
                    key_ini, bond_i = bond_ini
                    atoms_i = [ini.list_of_atoms[bond_i.atom1].pdb_atom_name,
                               ini.list_of_atoms[bond_i.atom2].pdb_atom_name]
                    grow_atomnames = [atm.pdb_atom_name for i, atm in grw.list_of_atoms.items()]
                    for atom_i in atoms_i:
                        if atom_i in atoms_g:
                            atoms_strip = [ai.strip("_") for ai in atoms_i]
                            if atm_to_repl in atoms_strip:
                                result = function(bond_i.eq_dist, bond_g.eq_dist)
                                bond_g.eq_dist = result


class ReduceLinearly(ReduceProperty):
    def __init__(self, template, lambda_to_reduce, template_core=None, atom_to_replace=None):
        ReduceProperty.__init__(self, template, lambda_to_reduce, template_core, atom_to_replace)

    def reduce_value(self, value):
        result = value * self.lambda_to_reduce
        return result
    
    def reduce_value_from_diference(self, value_init, value_grown):
        diff = value_grown - value_init
        result = (diff * self.lambda_to_reduce) + value_init
        return result
     

class ReduceExponentially(ReduceProperty):
    def __init__(self, template, lambda_to_reduce):
        ReduceProperty.__init__(self, template, lambda_to_reduce)

    def reduce_value(self, value):
        result = value * self.lambda_to_reduce ** 2
        return result


def file_to_list_of_lines(file_path):
    with open(file_path, "r") as template:
        content = template.readlines()
    return content


def get_string_from_line(line, index_initial, index_final):
    string = line[index_initial:index_final]
    return string.strip()


def find_equal_pdb_atom_names(template1, template2):
    pdb_atom_names_tmpl_1 = [template1.list_of_atoms[n].pdb_atom_name for n in range(1, len(template1.list_of_atoms)+1)]
    pdb_atom_names_tmpl_2 = [template2.list_of_atoms[n].pdb_atom_name for n in range(1, len(template2.list_of_atoms)+1)]
    return list(set(pdb_atom_names_tmpl_1).intersection(pdb_atom_names_tmpl_2))


def detect_atoms(template_initial, template_grown, hydrogen_to_replace):
    fragment_atoms = []
    core_atoms_in = []
    core_atoms_grown = []
    core_atoms_names = find_equal_pdb_atom_names(template_initial, template_grown)
    # First fill the list of the template grown
    for key, atom in template_grown.list_of_atoms.items():
        pdb_atom_name = atom.pdb_atom_name
        if pdb_atom_name not in core_atoms_names:
            fragment_atoms.append(atom)
        elif hydrogen_to_replace in pdb_atom_name:
            fragment_atoms.append(atom)
        elif pdb_atom_name in core_atoms_names:
            core_atoms_grown.append(atom)
    # Then the ones of the initial template
    for key, atom in template_initial.list_of_atoms.items():
        pdb_atom_name = atom.pdb_atom_name
        if pdb_atom_name in core_atoms_names:
            core_atoms_in.append(atom)
    
    return fragment_atoms, core_atoms_in, core_atoms_grown


def set_fragment_atoms(list_of_fragment_atoms):
    for atom in list_of_fragment_atoms:
        atom.is_fragment = True


def detect_fragment_bonds(list_of_fragment_atoms, template_grown):
    fragment_bonds = []
    fragment_indexes = []
    for atom in list_of_fragment_atoms:
        fragment_indexes.append(atom.atom_id)
    fragment_indexes = list(set(fragment_indexes))
    for key, bond in template_grown.list_of_bonds.items():
        if key[0] in fragment_indexes and key[1] in fragment_indexes:
            fragment_bonds.append(bond)
    return fragment_bonds


def set_linker_bond(template):
    for key, bond in template.list_of_bonds.items():
        if template.list_of_atoms[bond.atom1].is_linker or template.list_of_atoms[bond.atom2].is_linker: 
            #Only the core one will be detected, thus we must check if the bond has fragment atoms to select the linker
            if template.list_of_atoms[bond.atom1].is_fragment or template.list_of_atoms[bond.atom2].is_fragment: 
                bond.is_linker=True


def set_fragment_bonds(list_of_fragment_bonds):
    for bond in list_of_fragment_bonds:
        bond.is_fragment = True


def set_connecting_atom(template_grown, pdb_atom_name):
    for key, atom in template_grown.list_of_atoms.items():
        if pdb_atom_name in atom.pdb_atom_name:
            atom.is_linker = True


def reduce_fragment_parameters_linearly(template_object, lambda_to_reduce):
    reductor = ReduceLinearly(template_object, lambda_to_reduce)
    reductor.reduce_nbon_params(reductor.reduce_value)
    reductor.reduce_bond_eq_dist(reductor.reduce_value)


def modify_core_parameters_linearly(template_grow, lambda_to_reduce, template_core):
    reductor = ReduceLinearly(template_grow, lambda_to_reduce, template_core)
    reductor.modify_core_nbond_params(reductor.reduce_value_from_diference)
    reductor.modify_core_bond_eq_dist(reductor.reduce_value_from_diference)
    reductor.modify_core_theta(reductor.reduce_value_from_diference)
    reductor.modify_core_phis(reductor.reduce_value_from_diference)

def modify_linkers_parameters_linearly(template_grow, lambda_to_reduce, template_core, atom_to_replace):
    reductor = ReduceLinearly(template_grow, lambda_to_reduce, template_core, atom_to_replace)
    reductor.modify_linker_bond_eq_dist(reductor.reduce_value_from_diference)


def main(template_initial_path, template_grown_path, step, total_steps, hydrogen_to_replace, core_atom_linker,
         tmpl_out_path):
    """
    Module to modify templates, currently working in OPLS2005. This main function basically compares two templates;
    an initial and a grown one, extracting the atoms of the fragment (that have been grown). Then, it uses this data
    to modify Linearly different attributes of the template, particularly, sigmas, charges, bond equilibrium distance,
    and the radius non polar SGB from atoms and bonds of the fragment. This modification is performed according to a
    lambda parameter that is computed dividing the current step by the total number of steps. Finally, the template is
    modified and written again to an output file.
    :param template_initial_path: Path to an OPLS2005 template of the core ligand.
    :type template_initial_path: str
    :param template_grown_path: Path to an OPLS2005 template of the ligand with the fragment added to the core.
    :type template_grown_path: str
    :param step: Current step of the total steps.
    :type step: int
    :param total_steps: Total number of steps.
    :type total_steps: int
    :param hydrogen_to_replace: PDB atom name of the hydrogen that will be replaced for the linking atom of the fragment.
    :type hydrogen_to_replace: str
    :param core_atom_linker: PDB atom name of the core that is linking the fragment.
    :type core_atom_linker: str
    :param tmpl_out_path: Output path for the template modified.
    :type tmpl_out_path: str
    :return: None
    """
    lambda_to_reduce = float(step/(total_steps+1))
    templ_ini = TemplateOPLS2005(template_initial_path)
    for bond in templ_ini.list_of_bonds:
        key, bond_cont = bond
    templ_grw = TemplateOPLS2005(template_grown_path)
    fragment_atoms, core_atoms_in, core_atoms_grown = detect_atoms(template_initial=templ_ini, 
                                                                   template_grown=templ_grw,
                                                                   hydrogen_to_replace=hydrogen_to_replace)
    set_fragment_atoms(list_of_fragment_atoms=fragment_atoms)
    set_connecting_atom(template_grown=templ_grw, pdb_atom_name=core_atom_linker)
    fragment_bonds = detect_fragment_bonds(list_of_fragment_atoms=fragment_atoms, template_grown=templ_grw)
    set_fragment_bonds(list_of_fragment_bonds=fragment_bonds)
    set_linker_bond(templ_grw)
    modify_core_parameters_linearly(templ_grw, lambda_to_reduce, templ_ini)
    reduce_fragment_parameters_linearly(templ_grw, lambda_to_reduce)
    modify_linkers_parameters_linearly(templ_grw, lambda_to_reduce, templ_ini, hydrogen_to_replace)
    templ_grw.write_template_to_file(template_new_name=tmpl_out_path)






#Sebastian Raschka 2014
# Script that calculates the weighted center of mass of PDB files.
# Usage python3 center_of_mass.py file.pdb

# Atomic masses
# list of atomic weights from http://en.wikipedia.org/wiki/List_of_elements


ATOMIC_WEIGHTS = {'H': 1.008, 'HE': 4.002602, 'LI': 6.94, 'BE': 9.012182,
                  'B': 10.81, 'C': 12.011, 'N': 14.007, 'O': 15.999, 'F': 18.9984032,
                  'NE': 20.1797, 'NA': 22.98976928, 'MG': 24.305, 'AL': 26.9815386,
                  'SI': 28.085, 'P': 30.973762, 'S': 32.06, 'CL': 35.45, 'AR': 39.948,
                  'K': 39.0983, 'CA': 40.078, 'SC': 44.955912, 'TI': 47.867, 'V': 50.9415,
                  'CR': 51.9961, 'MN': 54.938045, 'FE': 55.845, 'CO': 58.933195,
                  'NI': 58.6934, 'CU': 63.546, 'ZN': 65.38, 'GA': 69.723, 'GE': 72.630,
                  'AS': 74.92160, 'SE': 78.96, 'BR': 79.904, 'RB': 85.4678, 'SR': 87.62,
                  'Y': 88.90585, 'ZR': 91.224, 'NB': 92.90638, 'MO': 95.96, 'TC': 98,
                  'RU': 101.07, 'RH': 102.90550, 'PD': 106.42, 'AG': 107.8682, 'CD': 112.411,
                  'IN': 114.818, 'SN': 118.710, 'SB': 121.760, 'TE': 127.60, 'I': 126.90447,
                  'XE': 131.293, 'CS': 132.9054519, 'BA': 137.327, 'LA': 138.90547,
                  'CE': 140.116, 'PR': 140.90765, 'ND': 144.242, 'PM': 145, 'SM': 150.36,
                  'EU': 151.964, 'GD': 157.25, 'TB': 158.92535, 'DY': 162.500, 'HO': 164.93032,
                  'ER': 167.259, 'TM': 168.93421, 'YB': 173.054, 'LU': 174.9668, 'HF': 178.49,
                  'TA': 180.94788, 'W': 183.84, 'RE': 186.207, 'OS': 190.23, 'IR': 192.217,
                  'PT': 195.084, 'AU': 196.966569, 'HG': 200.592, 'TL': 204.38, 'PB': 207.2,
                  'BI': 208.98040, 'PO': 209, 'AT': 210, 'RN': 222, 'FR': 223, 'RA': 226,
                  'AC': 227, 'TH': 232.03806, 'PA': 231.03588, 'U': 238.02891, 'NP': 237,
                  'PU': 244, 'AM': 243, 'CM': 247, 'BK': 247, 'CF': 251, 'ES': 252, 'FM': 257,
                  'MD': 258, 'NO': 259, 'LR': 262, 'RF': 267, 'DB': 268, 'SG': 269, 'BH': 270,
                  'HS': 269, 'MT': 278, 'DS': 281, 'RG': 281, 'CN': 285, 'UUT': 286, 'FL': 289,
                  'UUP': 288, 'LV': 293, 'UUS': 294}


def center_of_mass(pdbfile, include='ATOM,HETATM'):
  """
  Calculates center of mass of a protein and/or ligand structure.
  Returns:
      center (list): List of float coordinates [x,y,z] that represent the
      center of mass (precision 3).
  """

  center = [None, None, None]
  include = tuple(include.split(','))

  with open(pdbfile, 'r') as pdb:

    # extract coordinates [ [x1,y1,z1], [x2,y2,z2], ... ]
    coordinates = []
    masses = []
    for line in pdb:
      if line.startswith(include):
        coordinates.append([float(line[30:38]),    # x_coord
                            float(line[38:46]),    # y_coord
                            float(line[46:54])     # z_coord
                            ])
        element_name = line[76:].strip()
        if element_name not in ATOMIC_WEIGHTS:
          element_name = line.split()[2].strip()[0]
        masses.append(ATOMIC_WEIGHTS[element_name])

    assert len(coordinates) == len(masses)

    # calculate relative weight of every atomic mass
    total_mass = sum(masses)
    weights = [float(atom_mass / total_mass) for atom_mass in masses]

    # calculate center of mass
    center = [sum([coordinates[i][j] * weights[i]
                   for i in range(len(weights))]) for j in range(3)]
    center_rounded = [round(center[i], 3) for i in range(3)]
    return center_rounded


if __name__ == '__main__':

  import argparse

  parser = argparse.ArgumentParser(
      description='Calculates the weighted center of mass for structures in a PDB file.\n'
                  'By default, all atoms in the PDB file are included in the calculation.',
      epilog='Example:\n'
      'center_of_mass.py ~/Desktop/protein.pdb \n'
      '[-8.125, 20.461, -10.438]'
      '\n\nNote that for the center of mass calculation, the relative\natomic'
      ' weights are taken into account (atomic mass unit [u]).\n\n'
      'A list of the atomic weights can be found, e.g., at\n'
      'http://en.wikipedia.org/wiki/List_of_elements',
      formatter_class=argparse.RawTextHelpFormatter
  )

  parser.add_argument('PDBfile')
  parser.add_argument('-i', '--include', type=str,
                      default='ATOM,HETATM',
                      metavar='coordinate-ID',
                      help='Coordinate lines to include (default: "ATOM,HETATM")')

  parser.add_argument('-v', '--version', action='version', version='center_of_mass v. 1.0')

  args = parser.parse_args()

  center_of_mass = center_of_mass(args.PDBfile, include=args.include)

  for component in center_of_mass:
    print(component)



# General imports
import sys
import time
import glob
import argparse
import os
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), "AdaptivePELE_repo"))
import logging
from logging.config import fileConfig
import shutil
import subprocess
import traceback
# Local imports
from frag_pele.Helpers import clusterizer, checker, folder_handler, runner, constraints, check_constants
from frag_pele.Helpers import helpers, correct_fragment_names, center_of_mass
from frag_pele.Growing import template_fragmenter, simulations_linker
from frag_pele.Growing import add_fragment_from_pdbs, bestStructs
from frag_pele.Analysis import analyser
from frag_pele.Banner import Detector as dt
from frag_pele import serie_handler
import frag_pele.constants as c

# Calling configuration file for log system
FilePath = os.path.abspath(__file__)
PackagePath = os.path.dirname(FilePath)
LogPath = os.path.join(PackagePath, c.CONFIG_PATH)
fileConfig(LogPath)

# Getting the name of the module for the log system
logger = logging.getLogger(__name__)

# Get current path
curr_dir = os.path.abspath(os.path.curdir)


def parse_arguments():
    """
        Parse user arguments
        Output: list with all the user arguments
    """
    # All the docstrings are very provisional and some of them are old, they would be changed in further steps!!
    parser = argparse.ArgumentParser(description="""Description: FrAG is a Fragment-based ligand growing software which
    performs automatically the addition of several fragments to a core structure of the ligand in a protein-ligand
    complex.""")
    required_named = parser.add_argument_group('required named arguments')

    # FrAG related arguments
    required_named.add_argument("-cp", "--complex_pdb", required=True,
                        help="""Path to the PDB file which must contain a protein-ligand complex. Its ligand will be
                        used as the core structure. Remember to rename the ligand chain with a different character in
                        order to detect it.""")
    required_named.add_argument("-sef", "--serie_file", required=True,
                        help=""" Name of the tabular file which must contain the instructions required to perform several
                        successive growings, using different fragments or different growing positions.
                        
                        To do simple growings:
                        col1   col2    col3
                        To do successive growings:
                        (col1   col2    col3) x n_growings

                        Where col1 is the path to the PDB file of the fragment that will be added to the core structure
                        (name the chain of the fragment "L" by default).
                        Col2 is a string with the PDB atom name of the heavy atom of the core (the ligand contained in
                        the complex) where you would like to start the growing and create a new bond with the fragment.
                        And col3 is a string with the PDB atom name of the heavy atom of the fragment that will be used
                        to perform the bonding with the core.
                        """)
    parser.add_argument("--debug", action="store_true", help="Run Frag without launching PELE simulation")
    parser.add_argument("-nc", "--no_check", action="store_true", help="Don't perform the environment variables check")
    parser.add_argument("-x", "--growing_steps", type=int, default=c.GROWING_STEPS,
                        help="""Number of Growing Steps (GS). By default = {}.""".format(c.GROWING_STEPS))
    parser.add_argument("-cr", "--criteria", default=c.SELECTION_CRITERIA,
                        help="""Name of the column of the report file to select the structures that will spawn in the 
                        next GS. Additionally, this parameter will be the selection criteria to extract the best 
                        structure after completing the growing. By default = {}.""".format(c.SELECTION_CRITERIA))
    parser.add_argument("-rst", "--restart", action="store_true",
                        help="If set FrAG will continue from the last GS detected. If all GS are finished it will"
                             "restart in the equilibration phase.")
    parser.add_argument("-cc", "--c_chain", default="L", help="Chain name of the core. By default = 'L'")
    parser.add_argument("-fc", "--f_chain", default="L", help="Chain name of the fragment. By default = 'L'")
    parser.add_argument("-tc", "--clash_thr", default=1.7, help="Threshold distance that would to classify intramolecular"
                                                                "clashes.")
    parser.add_argument("-sc",  "--sampling_control", default=None, help="If set, templatized control file to use in the"
                                                                         " sampling simulation.")
    parser.add_argument("-op",  "--only_prepare", action="store_true", help="If set, all files to run growing are"
                                                                            " prepared, it stops before running PELE.")
    parser.add_argument("-og", "--only_grow", action="store_true", help="If set, it runs all growings of folders "
                                                                        "already prepared.")

    # Plop related arguments
    parser.add_argument("-pl", "--plop_path", default=c.PLOP_PATH,
                        help="Absolute path to PlopRotTemp.py. By default = {}".format(c.PLOP_PATH))
    parser.add_argument("-sp", "--sch_python", default=c.SCHRODINGER_PY_PATH,
                        help="""Absolute path to Schrodinger's python. 
                        By default = {}""".format(c.SCHRODINGER_PY_PATH))
    parser.add_argument("-rot", "--rotamers", default=c.ROTRES, type=str,
                        help="""Rotamers threshold used in the rotamers' library. 
                            By default = {}""".format(c.ROTRES))


    # PELE configuration arguments
    parser.add_argument("-d", "--pele_dir", default=c.PATH_TO_PELE,
                        help="Complete path to Pele_serial. "
                             "By default = {}".format(c.PATH_TO_PELE))
    parser.add_argument("-c", "--contrl", default=c.CONTROL_TEMPLATE,
                        help="Path to PELE's control file templatized. By default = {}".format(c.CONTROL_TEMPLATE))
    parser.add_argument("-l", "--license", default=c.PATH_TO_LICENSE,
                        help="Absolute path to PELE's licenses folder. "
                             " By default = {}".format(c.PATH_TO_LICENSE))
    parser.add_argument("-r", "--resfold", default=c.RESULTS_FOLDER,
                        help="Name for PELE's results folder. By default = {}".format(c.RESULTS_FOLDER))
    parser.add_argument("-rp", "--report", default=c.REPORT_NAME,
                        help="Suffix name of the report file from PELE. By default = {}".format(c.REPORT_NAME))
    parser.add_argument("-tj", "--traject", default=c.TRAJECTORY_NAME,
                        help="Suffix name of the trajectory file from PELE. By default = {}".format(c.TRAJECTORY_NAME))
    parser.add_argument("-cs", "--cpus", type=int, default=c.CPUS,
                        help="Number of cores (computational power) to paralellize PELE's simulations."
                             "By default = {}".format(c.CPUS))
    parser.add_argument("-stp", "--steps", type=int, default=c.STEPS,
                        help="""Number of simulation steps inside each GS. By default = {}""".format(c.STEPS))
    parser.add_argument("-es", "--pele_eq_steps", default=c.PELE_EQ_STEPS,
                        help="Number of PELE steps in equilibration. By default = {}".format(c.PELE_EQ_STEPS))
    parser.add_argument("-miov", "--min_overlap", default=c.MIN_OVERLAP,
                        help="Minimum value of overlapping factor used in the control_file of PELE. "
                             "By default = {}".format(c.MIN_OVERLAP))
    parser.add_argument("-maov", "--max_overlap", default=c.MAX_OVERLAP,
                        help="Maximum value of overlapping factor used in the control_file of PELE."
                             " By default = {}".format(c.MAX_OVERLAP))
    parser.add_argument("-tmp", "--temperature", default=c.TEMPERATURE,
                        help="Temperature value to add in the control file. If the temperature is high more steps of "
                             "PELE will be accepted when applying the Metropolis Criteria. "
                             "By default = {}".format(c.TEMPERATURE))
    parser.add_argument("-sd", "--seed", default=c.SEED,
                        help="Seed to get the random numbers that will be used in the PELE simulation"
                             "By default = {}".format(c.SEED))
    parser.add_argument("-st", "--steering", default=c.STEERING,
                        help="Steering that will be used in the PELE simulation"
                             "By default = {}".format(c.STEERING))
    parser.add_argument("-trh", "--translation_high", default=c.TRANSLATION_HIGH,
                        help="Translation that will be used in the PELE simulation to displace the ligand. Low value."
                             "By default = {}".format(c.TRANSLATION_HIGH))
    parser.add_argument("-roth", "--rotation_high", default=c.ROTATION_HIGH,
                        help="Rad of rotation that will be used in the PELE simulation to perturb the ligand. "
                             "Low value. By default = {}".format(c.ROTATION_HIGH))
    parser.add_argument("-trl", "--translation_low", default=c.TRANSLATION_LOW,
                        help="Translation that will be used in the PELE simulation to displace the ligand. Low value."
                             "By default = {}".format(c.TRANSLATION_LOW))
    parser.add_argument("-rotl", "--rotation_low", default=c.ROTATION_LOW,
                        help="Rad of rotation that will be used in the PELE simulation to perturb the ligand. "
                             "Low value. By default = {}".format(c.ROTATION_LOW))
    parser.add_argument("-rad", "--radius_box", default=c.RADIUS_BOX,
                        help="Size of the radius to define the box in the PELE simulation where the ligand will be"
                             "perturbed. By default = {}".format(c.RADIUS_BOX))
    parser.add_argument("-dat", "--data", default=c.PATH_TO_PELE_DATA,
                        help="Path to PELE Data folder.")
    parser.add_argument("-doc", "--documents", default=c.PATH_TO_PELE_DOCUMENTS,
                        help="Path to PELE Documents folder.")
    parser.add_argument("-sr", "--srun", default=True,
                        help="If true it runs PELE with srun command, else it will run it with mpirun.")

    # Clustering related arguments
    parser.add_argument("-dis", "--distcont", default=c.DISTANCE_COUNTER,
                        help="""Distance used to determine which amino acids are in contact with the ligand to generate 
                        different clusters of structures to initialize the next GS.
                        By default = {}""".format(c.DISTANCE_COUNTER))
    parser.add_argument("-ct", "--threshold", default=c.CONTACT_THRESHOLD,
                        help="Threshold distance used in the clustering. By default = {}".format(c.CONTACT_THRESHOLD))
    parser.add_argument("-e", "--epsilon", default=c.EPSILON,
                        help="An epsilon fraction of processors are distributed proportionally to the value of a metric,"
                             "and the rest are inverselyProportional distributed. A param n can be specified to only "
                             "consider the n clusters with best metric. By default = {}".format(c.EPSILON))
    parser.add_argument("-cn", "--condition", default=c.CONDITION,
                        help="Selects wether to take into account maximum or minimum values in epsilon related spawning,"
                             "values are min or max. By default = {}".format(c.CONDITION))
    parser.add_argument("-mw", "--metricweights", default=c.METRICS_WEIGHTS,
                        help="Selects how to distribute the weights of the cluster according to its metric, "
                             "two options: linear (proportional to metric) or Boltzmann weigths (proportional "
                             "to exp(-metric/T). Needs to define the temperature T. "
                             "By default = {}".format(c.METRICS_WEIGHTS))
    parser.add_argument("-ncl", "--nclusters", default=c.NUM_CLUSTERS,
                        help="Number of initial structures that we want to use in each new GS. "
                             "By default = {}".format(c.NUM_CLUSTERS))

    parser.add_argument("-pdbf", "--pdbout", default=c.PDBS_OUTPUT_FOLDER,
                        help="Folder where PDBs selected to spawn in the next GS will be stored."
                             "By default = {}".format(c.PDBS_OUTPUT_FOLDER))
    parser.add_argument("-ban", "--banned", default=c.BANNED_DIHEDRALS_ATOMS, type=str, nargs='+', action='append',
                        help="List of lists of quartets of PDB atom names that form the banned dihedrals."
                             "By default = {}".format(c.BANNED_DIHEDRALS_ATOMS))
    parser.add_argument("-lim", "--limit", default=c.BANNED_ANGLE_THRESHOLD, type=int,
                        help="Maximum degrees that can accept the banned dihedral."
                             "By default = {}".format(c.BANNED_ANGLE_THRESHOLD))

    #Protocol argument
    parser.add_argument("-HT", "--highthroughput", action="store_true",
                        help="Run frag pele high-throughput mode")

    parser.add_argument("-EX", "--explorative", action="store_true",
                        help="Run frag pele explorative mode: sampling simulation with high movement of the ligand.")

    parser.add_argument("--test", action="store_true", help="run test config")

    #Output format option
    parser.add_argument("--mae", action="store_true",
                        help="Retrieve .mae files intead of pdbs")

    #Others
    parser.add_argument("--rename", action="store_true",
                        help="Avoid core renaming")

    args = parser.parse_args()


    return args.complex_pdb, args.growing_steps, \
           args.criteria, args.plop_path, args.sch_python, args.pele_dir, args.contrl, args.license, \
           args.resfold, args.report, args.traject, args.pdbout, args.cpus, \
           args.distcont, args.threshold, args.epsilon, args.condition, args.metricweights, args.nclusters, \
           args.pele_eq_steps, args.restart, args.min_overlap, args.max_overlap, args.serie_file, \
           args.c_chain, args.f_chain, args.steps, args.temperature, args.seed, args.rotamers, \
           args.banned, args.limit, args.mae, args.rename, args.clash_thr, args.steering, \
           args.translation_high, args.rotation_high, args.translation_low, args.rotation_low, args.explorative, \
           args.radius_box, args.sampling_control, args.data, args.documents, args.only_prepare, args.only_grow, \
           args.no_check, args.debug, args.highthroughput, args.test, args.srun


def grow_fragment(complex_pdb, fragment_pdb, core_atom, fragment_atom, iterations, criteria, plop_path, sch_python,
                  pele_dir, contrl, license, resfold, report, traject, pdbout, cpus, distance_contact, clusterThreshold,
                  epsilon, condition, metricweights, nclusters, pele_eq_steps, restart, min_overlap, max_overlap, ID,
                  h_core=None, h_frag=None, c_chain="L", f_chain="L", steps=6, temperature=1000, seed=1279183, rotamers="30.0",
                  banned=None, limit=None, mae=False, rename=False, threshold_clash=1.7, steering=0,
                  translation_high=0.05, rotation_high=0.10, translation_low=0.02, rotation_low=0.05, explorative=False,
                  radius_box=4, sampling_control=None, data=None, documents=None, only_prepare=False, only_grow=False, 
                  no_check=False, debug=False, srun=True):
    """
    Description: FrAG is a Fragment-based ligand growing software which performs automatically the addition of several
    fragments to a core structure of the ligand in a protein-ligand complex.
    :param complex_pdb: Path to the PDB file which must contain a protein-ligand complex. Its ligand will be
    used as the core structure. Remember to rename the ligand chain with a different character in
    order to detect it.
    :type complex_pdb: str
    :param fragment_pdb: Path to the PDB file containing the fragment.
    :type fragment_pdb: str
    :param core_atom: PDB-atom-name of the atom of the core that will be used as starting point to grow the fragment.
    :type core_atom: str (max length: 4)
    :param fragment_atom: PDB-atom-name of the atom of the fragment that will bond the core.
    :type fragment_atom: str (max length: 4)
    :param iterations: Number of Growing Steps (GS).
    :type iterations: int
    :param criteria: Name of the column of the report file to select the structures that will spawn in the
    next GS. Additionally, this parameter will be the selection criteria to extract the best
    structure after completing the growing.
    :type criteria: str
    :param plop_path: Absolute path to PlopRotTemp.py.
    :type plop_path: str
    :param sch_python: Absolute path to Schrodinger's python.
    :type sch_python: str
    :param pele_dir: Absolute path to PELE executables.
    :type pele_dir: str
    :param contrl: Name of the PELE's control file templatized.
    :type contrl: str
    :param license: Absolute path to PELE's license file.
    :type license: str
    :param resfold: Name of the results folder.
    :type resfold: str
    :param report: Prefix name of PELE's output file.
    :type report: str
    :param traject: Prefix name of the trajectory file generated by PELE.
    :type traject: str
    :param pdbout: Prefix name of the output PDB extracted from FrAG in each iteration.
    :type pdbout: str
    :param cpus: Number of CPUs that will be used to perform the simulation.
    :type cpus: int
    :param distance_contact: This distance will be used to determine which amino acids are in contact with the ligand.
    Then, this information will be useful to generate different clusters of structures to initialize the next iteration.
    :type distance_contact: float
    :param clusterThreshold: Threshold that will be used in the clustering step.
    :type clusterThreshold: float (from 0 to 1)
    :param epsilon: Epsilon parameter used in the clustering.
    :type epsilon: float (from 0 to 1)
    :param condition: Condition to select best values to cluster: minimum or maximum.
    :type condition: str
    :param metricweights: Selects how to distribute the weights of the cluster according to its metric,
    two options: linear (proportional to metric) or Boltzmann weigths (proportional to exp(-metric/T).
    Needs to define the temperature T.
    :type metricweights: str
    :param nclusters: Number of cluster that we would like to generate.
    :type nclusters: int
    :param pele_eq_steps: Number of PELE steps that we want to perform in the equilibration.
    :type pele_eq_steps: int
    :param restart: If set FrAG will find your last iteration completed and will restart from this point.
    :type restart: bool
    :param min_overlap: Minimun value of overlapping factor that will be replaced in the control file.
    :type min_overlap: float (from 0 to 1)
    :param max_overlap: Maximum value of overlapping factor that will be replaced in the control file.
    :type max_overlap: float (from 0 to 1)
    :param ID: Name to identify the new ligand in certain folder.
    :type ID: str
    :param h_core: PDB-atom-name of the hydrogen bonded to the selected heavy atom of the core that will be replaced for
     the fragment, being the initial point of the growing.
    :type h_core: str (max length: 4)
    :param h_frag: PDB-atom-name of the hydrogen bonded to the selected heavy atom of the fragment that will removed to
    bond the fragment with the core.
    :type h_frag: str (max length: 4)
    :param c_chain: Chain name for the ligand in the complex_pdb.
    :type c_chain: str (max length: 1)
    :param f_chain: Chain name for the ligand in the fragment_pdb.
    :type f_chain: str (max length: 1)
    :param steps: PELE steps to do in each GS.
    :type steps: int
    :param temperature: Temperature to add in the control file.
    :type temperature: int
    :param seed: Seed in the PELE's control file.
    :type seed: int
    :param rotamers: Degrees of rotation used to select the rotamer's library. Higher value, faster but less accuracy.
    :type rotamers: str
    :param banned: If set, list of tuples with the atom names of the angles that will not be modified.
    :type banned: list
    :param limit: Limit angle in degrees of the banned angles. Any structure with a higher value will be discarted.
    :type limit: int
    :param mae: If set, the output structures will be saved in "mae." format.
    :type mae: bool
    :param rename: If set, the pdb atom names of the fragment will be renamed with "G+atom_number".
    :type rename: bool
    :param threshold_clash: Distance that will be used to detect contacts between atoms of the fragment and atoms
    of the core.
    :type threshold_clash: float
    :param steering: Steering parameter of the PELE's control file.
    :type steering: int
    :param translation_high: Translation (high value) of the PELE's control file. In anstrongs.
    :type translation_high: float
    :param translation_low: Translation (low value) of the PELE's control file. In anstrongs.
    :type translation_low: float
    :param rotation_high: Rotation (high value) of PELE's control file. In rad.
    :type rotation_high: float
    :param rotation_low: Rotation (low value) of PELE's control file. In rad.
    :type rotation_low: float
    :param explorative: Set all parameters to perform a more explorative simulation after the growing, in the sampling
    simulation.
    :type explorative: bool
    :param radius_box: Radius box size of PELE's control file. In anstrongs.
    :type radius_box: float
    :param sampling_control: templatized control file to be used in the sampling simulation.
    :type sampling_control: str
    :return:
    """
    #Check harcoded path in constants.py
    if not no_check:
        check_constants.check()
    # Time computations
    start_time = time.time()
    # Global variable to keep info
    simulation_info = []
    # Path definition
    plop_relative_path = os.path.join(PackagePath, plop_path)
    current_path = os.path.abspath(".")

    pdb_basename = complex_pdb.split(".pdb")[0]  # Get the name of the pdb without extension
    if "/" in pdb_basename:
        pdb_basename = pdb_basename.split("/")[-1]  # And if it is a path, get only the name

    working_dir = os.path.join(current_path, "{}_{}".format(pdb_basename, ID))
    if not os.path.exists(working_dir):
        os.mkdir(working_dir)  # Creating a working directory for each PDB-fragment combination
    pdbout_folder = os.path.join(working_dir, pdbout)
    path_to_templates_generated = os.path.join(working_dir,
                                               "DataLocal/Templates/OPLS2005/HeteroAtoms/templates_generated")
    path_to_templates = os.path.join(working_dir, "DataLocal/Templates/OPLS2005/HeteroAtoms")
    path_to_lib = os.path.join(working_dir, "DataLocal/LigandRotamerLibs")
    # Creation of output folder
    folder_handler.check_and_create_DataLocal(working_dir=working_dir)
    # Creating constraints
    const = "\n".join(constraints.retrieve_constraints(complex_pdb, {}, {}, 5, 5, 10))
    # Creating symbolic links
    helpers.create_symlinks(data, os.path.join(working_dir, 'Data'))
    helpers.create_symlinks(documents, os.path.join(working_dir, 'Documents'))
    #  ---------------------------------------Pre-growing part - PREPARATION -------------------------------------------
    fragment_names_dict, hydrogen_atoms, pdb_to_initial_template, pdb_to_final_template, pdb_initialize, \
    core_original_atom, fragment_original_atom = add_fragment_from_pdbs.main(complex_pdb, fragment_pdb, core_atom,
                                                                             fragment_atom, iterations, h_core=h_core,
                                                                             h_frag=h_frag, core_chain=c_chain,
                                                                             fragment_chain=f_chain, rename=rename,
                                                                             threshold_clash=threshold_clash,
                                                                             output_path=working_dir,
                                                                             only_grow=only_grow)

    # Create the templates for the initial and final structures
    template_resnames = []
    for pdb_to_template in [pdb_to_initial_template, pdb_to_final_template]:
        if not only_grow and not restart:
            cmd = "{} {} {} {} {} {}".format(sch_python, plop_relative_path, os.path.join(working_dir,
                                             add_fragment_from_pdbs.c.PRE_WORKING_DIR, pdb_to_template), rotamers,
                                             path_to_templates_generated, path_to_lib)

            try:
                subprocess.call(cmd.split())
            except OSError:
                raise OSError("Path {} not foud. Change schrodinger path under frag_pele/constants.py".format(sch_python))
        template_resname = add_fragment_from_pdbs.extract_heteroatoms_pdbs(os.path.join(working_dir, add_fragment_from_pdbs.
                                                                           c.PRE_WORKING_DIR, pdb_to_template),
                                                                           False, c_chain, f_chain)
        template_resnames.append(template_resname)

    # Set box center from ligand COM
    resname_core = template_resnames[0]
    center = center_of_mass.center_of_mass(os.path.join(working_dir, c.PRE_WORKING_DIR, "{}.pdb".format(resname_core)))

    # Get template filenames
    template_initial, template_final = ["{}z".format(resname.lower()) for resname in template_resnames]

    if only_prepare:
        print("Files of {} prepared".format(ID))
        return

    # --------------------------------------------GROWING SECTION-------------------------------------------------------
    # Lists definitions

    templates = ["{}_{}".format(os.path.join(path_to_templates_generated, template_final), n) for n in range(0, iterations+1)]

    results = [os.path.join(working_dir, c.OUTPUT_FOLDER, str(n)) for n in range(0, iterations+1)]

    pdbs = [pdb_initialize if n == 0 else "{}_{}".format(n, pdb_initialize) for n in range(0, iterations+1)]

    pdb_selected_names = ["initial_0_{}.pdb".format(n) for n in range(0, cpus-1)]

    # Generate starting templates
    template_fragmenter.main(template_initial_path=os.path.join(path_to_templates_generated, template_initial),
                             template_grown_path=os.path.join(path_to_templates_generated, template_final),
                             step=1, total_steps=iterations, hydrogen_to_replace=core_original_atom,
                             core_atom_linker=core_atom,
                             tmpl_out_path=os.path.join(path_to_templates_generated, "{}_0".format(template_final)))

    # Make a copy in the main folder of Templates in order to use it as template for the simulation
    shutil.copy(os.path.join(path_to_templates_generated, "{}_0".format(template_final)),
                os.path.join(path_to_templates, template_final))  # Replace the original template in the folder

    # Clear PDBs folder
    if not restart:
        list_of_subfolders = glob.glob("{}*".format(pdbout_folder))
        for subfolder in list_of_subfolders:
            shutil.rmtree(subfolder)

    # Simulation loop - LOOP CORE
    for i, (template, pdb_file, result) in enumerate(zip(templates, pdbs, results)):

        # Only if reset
        if restart:
            if os.path.exists(os.path.join(pdbout_folder, "{}".format(i))) and os.path.exists(os.path.join(pdbout_folder,
                                                                                                    "{}".format(i),
                                                                                                    "initial_0_0.pdb")):
                print("STEP {} ALREADY DONE, JUMPING TO THE NEXT STEP...".format(i))
                continue
        # Otherwise start from the beggining
        pdb_input_paths = ["{}".format(os.path.join(pdbout_folder, str(i-1), pdb_file)) for pdb_file in pdb_selected_names]
        # Banned dihedrals will be checked here
        if banned:
            pdbs_with_banned_dihedrals = Detector.check_folder(folder=os.path.join(pdbout_folder, str(i-1)),
                                                               threshold=limit,
                                                               dihedrals=banned,
                                                               lig_chain=c_chain,
                                                               processors=cpus)
            pdb_input_paths = [pdb_file for pdb_file, flag in pdbs_with_banned_dihedrals.items() if flag]

        # Control file modification
        overlapping_factor = float(min_overlap) + (((float(max_overlap) - float(min_overlap))*i) / iterations)
        overlapping_factor = "{0:.2f}".format(overlapping_factor)

        if i != 0:
            # Check atom overlapping
            pdbs_with_overlapping = clusterizer.check_atom_overlapping(pdb_input_paths)
            pdb_input_paths_checked = []
            for pdb in pdb_input_paths:
                if pdb not in pdbs_with_overlapping:
                    pdb_input_paths_checked.append(pdb)
            simulation_file = simulations_linker.control_file_modifier(contrl, pdb=pdb_input_paths_checked, step=i,
                                                                       license=license,
                                                                       working_dir=working_dir,
                                                                       overlap=overlapping_factor, results_path=result,
                                                                       steps=steps,
                                                                       chain=c_chain, constraints=const, center=center,
                                                                       temperature=temperature, seed=seed,
                                                                       steering=steering,
                                                                       translation_high=translation_high,
                                                                       translation_low=translation_low,
                                                                       rotation_high=rotation_high,
                                                                       rotation_low=rotation_low,
                                                                       radius=radius_box)
        else:
            logger.info(c.SELECTED_MESSAGE.format(contrl, pdb_initialize, result, i))
            simulation_file = simulations_linker.control_file_modifier(contrl, pdb=[pdb_initialize], step=i,
                                                                       license=license,
                                                                       working_dir=working_dir,
                                                                       overlap=overlapping_factor, results_path=result,
                                                                       steps=steps,
                                                                       chain=c_chain, constraints=const, center=center,
                                                                       temperature=temperature, seed=seed,
                                                                       steering=steering,
                                                                       translation_high=translation_high,
                                                                       translation_low=translation_low,
                                                                       rotation_high=rotation_high,
                                                                       rotation_low=rotation_low,
                                                                       radius=radius_box)

        logger.info(c.LINES_MESSAGE)
        if i != 0:
            template_fragmenter.main(template_initial_path=os.path.join(path_to_templates_generated, template_initial),
                                     template_grown_path=os.path.join(path_to_templates_generated, template_final),
                                     step=i+1, total_steps=iterations, hydrogen_to_replace=core_original_atom,
                                     core_atom_linker=core_atom,
                                     tmpl_out_path=os.path.join(path_to_templates, template_final))

        # Make a copy of the template file in growing_templates folder
        shutil.copy(os.path.join(path_to_templates, template_final), template)

        # Creating results folder
        folder_handler.check_and_create_results_folder(result, working_dir)
        # ------SIMULATION PART------
        # Change directory to the working one
        os.chdir(working_dir)
        if debug:
            return 
        else:
            simulations_linker.simulation_runner(pele_dir, simulation_file, cpus, srun)
        logger.info(c.LINES_MESSAGE)
        logger.info(c.FINISH_SIM_MESSAGE.format(result))
        # Before selecting a step from a trajectory we will save the input PDB file in a folder
        folder_handler.check_and_create_pdb_clusters_folder(pdbout_folder, i)

        # ---------------------------------------------------CLUSTERING-------------------------------------------------
        # Transform column name of the criteria to column number
        result_abs = os.path.abspath(result)
        logger.info("Looking structures to cluster in '{}'".format(result_abs))
        column_number = clusterizer.get_column_num(result_abs, criteria, report)
        # Selection of the trajectory used as new input
        clusterizer.cluster_traject(str(template_resnames[1]), cpus-1, column_number, distance_contact,
                                        clusterThreshold, "{}*".format(os.path.join(result_abs, traject)),
                                        os.path.join(pdbout_folder, str(i)), os.path.join(result_abs),
                                        epsilon, report, condition, metricweights, nclusters)
    # ----------------------------------------------------EQUILIBRATION-------------------------------------------------
    # Set input PDBs
    pdb_inputs = ["{}".format(os.path.join(pdbout_folder, str(iterations), pdb_file)) for pdb_file in pdb_selected_names]
    if banned:
        pdbs_with_banned_dihedrals = Detector.check_folder(folder=os.path.join(pdbout_folder, str(iterations)),
                                                           threshold=limit,
                                                           dihedrals=banned,
                                                           lig_chain=c_chain,
                                                           processors=cpus)
        pdb_inputs = [pdb_file for pdb_file, flag in pdbs_with_banned_dihedrals.items() if flag]
    if not os.path.exists(os.path.join(working_dir, "sampling_result")):  # Create the folder if it does not exist
        os.mkdir(os.path.join(working_dir, "sampling_result"))
    # Modify the control file to increase the steps TO THE SAMPLING SIMULATION
    if sampling_control:
        simulation_file = simulations_linker.control_file_modifier(sampling_control, pdb=pdb_inputs, step=iterations,
                                                                   license=license, working_dir=working_dir,
                                                                   overlap=max_overlap,
                                                                   results_path=os.path.join(working_dir, "sampling_result"),
                                                                   steps=pele_eq_steps, chain=c_chain,
                                                                   constraints=const, center=center,
                                                                   temperature=temperature, seed=seed, steering=steering,
                                                                   translation_high=translation_high,
                                                                   translation_low=translation_low,
                                                                   rotation_high=rotation_high, rotation_low=rotation_low,
                                                                   radius=radius_box)
    elif explorative and not sampling_control:
        simulation_file = simulations_linker.control_file_modifier(contrl, pdb=pdb_inputs, license=license,
                                                                   working_dir=working_dir, step=iterations,
                                                                   overlap=max_overlap,
                                                                   results_path=os.path.join(working_dir, "sampling_result"),
                                                                   steps=pele_eq_steps,
                                                                   chain=c_chain, constraints=const, center=center,
                                                                   temperature=temperature, seed=seed,
                                                                   steering=2,
                                                                   translation_high=0.5,
                                                                   translation_low=0.3,
                                                                   rotation_high=0.4,
                                                                   rotation_low=0.15,
                                                                   radius=25)
    else:
        simulation_file = simulations_linker.control_file_modifier(contrl, pdb=pdb_inputs, step=iterations,
                                                                   license=license, overlap=max_overlap,
                                                                   working_dir=working_dir,
                                                                   results_path=os.path.join(working_dir, "sampling_result"),
                                                                   steps=pele_eq_steps, chain=c_chain,
                                                                   constraints=const, center=center,
                                                                   temperature=temperature, seed=seed, steering=steering,
                                                                   translation_high=translation_high,
                                                                   translation_low=translation_low,
                                                                   rotation_high=rotation_high, rotation_low=rotation_low,
                                                                   radius=radius_box)

    # EQUILIBRATION SIMULATION
    # Change directory to the working one
    os.chdir(working_dir)
    shutil.copy(os.path.join(path_to_templates_generated, template_final), path_to_templates)
    if not (restart and os.path.exists("top_result")):
        logger.info(".....STARTING EQUILIBRATION.....")
        simulations_linker.simulation_runner(pele_dir, simulation_file, cpus, srun)
    os.chdir(curr_dir)
    equilibration_path = os.path.join(working_dir, "sampling_result")
    # SELECTION OF BEST STRUCTURES
    selected_results_path = os.path.join(working_dir, "top_result")
    if not os.path.exists(selected_results_path):  # Create the folder if it does not exist
        os.mkdir(selected_results_path)
    best_structure_file, all_output_files = bestStructs.main(criteria, selected_results_path, path=equilibration_path,
                                                             n_structs=50)

    shutil.copy(os.path.join(selected_results_path, best_structure_file), os.path.join(working_dir, c.PRE_WORKING_DIR,
                                                                                       selected_results_path + ".pdb"))
    # COMPUTE AND SAVE THE SCORE
    analyser.analyse_at_epoch(report_prefix=report, path_to_equilibration=equilibration_path, execution_dir=curr_dir,
                              column=criteria, quantile_value=0.25)

    
    #MOVE FROM PDB TO MAE
    if mae:
        if sch_python.endswith("python"):
            schrodinger_path = os.path.dirname(os.path.dirname(sch_python))
        elif sch_python.endswith("run"):
            schrodinger_path = os.path.dirname(sch_python)
        python_file = os.path.join(os.path.dirname(FilePath), "Analysis/output_files.py")
        for outputfile in all_output_files:
            filename = os.path.join(selected_results_path, outputfile)
            command = "{} {} {} --schr {} {}".format(sch_python, python_file, filename, schrodinger_path, "--remove") 
            subprocess.call(command.split())

    # COMPUTE TIME
    end_time = time.time()
    total_time = (end_time - start_time) / 60
    logging.info("Growing of {} in {} min".format(fragment_pdb, total_time))

    return fragment_names_dict
    

def main(complex_pdb, serie_file, iterations=c.GROWING_STEPS, criteria=c.SELECTION_CRITERIA, plop_path=c.PLOP_PATH, sch_python=c.SCHRODINGER_PY_PATH, pele_dir=c.PATH_TO_PELE, contrl=c.CONTROL_TEMPLATE, license=c.PATH_TO_LICENSE, resfold=c.RESULTS_FOLDER, 
    report=c.REPORT_NAME, traject=c.TRAJECTORY_NAME, pdbout=c.PDBS_OUTPUT_FOLDER, cpus=c.CPUS, distcont=c.DISTANCE_COUNTER, threshold=c.CONTACT_THRESHOLD, epsilon=c.EPSILON, condition=c.CONDITION, metricweights=c.METRICS_WEIGHTS, 
    nclusters=c.NUM_CLUSTERS, pele_eq_steps=c.PELE_EQ_STEPS, restart=False, min_overlap=c.MIN_OVERLAP, max_overlap=c.MAX_OVERLAP,
    c_chain="L", f_chain="L", steps=c.STEPS, temperature=c.TEMPERATURE, seed=c.SEED, rotamers=c.ROTRES, banned=c.BANNED_DIHEDRALS_ATOMS, limit=c.BANNED_ANGLE_THRESHOLD, mae=False,
    rename=None, threshold_clash=1.7, steering=c.STEERING, translation_high=c.TRANSLATION_HIGH, rotation_high=c.ROTATION_HIGH, 
    translation_low=c.TRANSLATION_LOW, rotation_low=c.ROTATION_LOW, explorative=False, radius_box=c.RADIUS_BOX, sampling_control=None, data=c.PATH_TO_PELE_DATA, documents=c.PATH_TO_PELE_DOCUMENTS, 
    only_prepare=False, only_grow=False, no_check=False, debug=False, protocol=False, test=False, srun=True):

    if protocol == "HT":
        iteration = 1
        steps = 3
        pele_eq_steps = 10

    if test:
        iteration = 1
        steps = 1
        steps = 1
        pele_eq_steps = 1
        temp = 1000000

    #HOT FIX!! Fix it properly
    original_dir = os.path.abspath(os.getcwd())
    list_of_instructions = serie_handler.read_instructions_from_file(serie_file)
    print("READING INSTRUCTIONS... You will perform the growing of {} fragments. GOOD LUCK and ENJOY the "
          "trip :)".format(len(list_of_instructions)))
    dict_traceback = correct_fragment_names.main(complex_pdb)
    for instruction in list_of_instructions:
        # We will iterate trough all individual instructions of file.
        os.chdir(original_dir)
        # SUCCESSIVE GROWING
        if type(instruction) == list:  #  If in the individual instruction we have more than one command means successive growing.
            growing_counter = len(instruction)  #  Doing so we will determinate how many successive growings the user wants to do.
            atomname_mappig = []
            for i in range(int(growing_counter)):
                core_from_previous_fragment = instruction[i][-1]
                fragment_pdb, core_atom, fragment_atom = instruction[i][0], instruction[i][1], instruction[i][2]
                atoms_if_bond = serie_handler.extract_hydrogens_from_instructions([fragment_pdb, core_atom, fragment_atom])
                if atoms_if_bond:
                    core_atom = atoms_if_bond[0]
                    if core_from_previous_fragment and i!=0:
                        previous_fragment_atomnames_map = atomname_mappig[int(instruction[i][4])-1]
                        core_atom = previous_fragment_atomnames_map[core_atom]
                        h_core = previous_fragment_atomnames_map[atoms_if_bond[1]]
                    else:
                        h_core = atoms_if_bond[1]
                    fragment_atom = atoms_if_bond[2]
                    h_frag = atoms_if_bond[3]
                else:
                    if core_from_previous_fragment and i!=0:
                        previous_fragment_atomnames_map = atomname_mappig[int(instruction[i][4])-1]
                        core_atom = previous_fragment_atomnames_map[core_atom]
                    h_core = None
                    h_frag = None
                if i == 0:  # In the first iteration we will use the complex_pdb as input.
                    complex_sequential_pdb = complex_pdb
                    ID = instruction[i][3]
                else:  # If is not the first we will use as input the output of the previous iteration
                    pdb_basename = complex_pdb.split(".pdb")[0]  # Get the name of the pdb without extension
                    if "/" in pdb_basename:
                        pdb_basename = pdb_basename.split("/")[-1]  # And if it is a path, get only the name
                    working_dir = "{}_{}".format(pdb_basename, ID)
                    complex_sequential_pdb = os.path.join(working_dir, "top_result.pdb")
                    dict_traceback = correct_fragment_names.main(complex_sequential_pdb)
                    ID_completed = []
                    for id in instruction[0:i+1]:
                        ID_completed.append(id[3])
                    ID = "".join(ID_completed)
                try:
                    ID = ID.split("/")[-1]
                except Exception:
                    os.chdir(original_dir)
                    traceback.print_exc()
                    if debug: raise Exception()
                try:
                    ID = ID.split("/")[-1]
                except Exception:
                    os.chdir(original_dir)
                    traceback.print_exc()
                    if debug: raise Exception()
                try:
                    serie_handler.check_instructions(instruction[i], complex_sequential_pdb, c_chain, f_chain)
                    print("PERFORMING SUCCESSIVE GROWING...")
                    print("HYDROGEN ATOMS IN INSTRUCTIONS:  {}    {}".format(h_core, h_frag))
                    atomname_map = grow_fragment(complex_sequential_pdb, fragment_pdb, core_atom, fragment_atom, 
                                   iterations, criteria, plop_path, sch_python,pele_dir, contrl, license, resfold, 
                                   report, traject, pdbout, cpus, distcont, threshold, epsilon, condition, 
                                   metricweights, nclusters, pele_eq_steps, restart, min_overlap, max_overlap, 
                                   ID, h_core, h_frag, c_chain, f_chain, steps, temperature, seed, rotamers, banned,
                                   limit, mae, rename, threshold_clash, steering, translation_high, rotation_high,
                                   translation_low, rotation_low, explorative, radius_box, sampling_control, data, documents,
                                   only_prepare, only_grow, no_check, debug, srun)
                    atomname_mappig.append(atomname_map)
 
                except Exception:
                    os.chdir(original_dir)
                    traceback.print_exc()
                    if debug: raise Exception()
        # INDIVIDUAL GROWING
        else:
            # Initialize the growing for each line in the file
            fragment_pdb, core_atom, fragment_atom, ID = instruction[0], instruction[1], instruction[2], instruction[3]
            atoms_if_bond = serie_handler.extract_hydrogens_from_instructions([fragment_pdb, core_atom, fragment_atom])
            try:
                complex_pdb = complex_pdb
                ID = ID.split("/")[-1]
            except Exception:
                os.chdir(original_dir)
                traceback.print_exc()
                if debug: raise Exception()
 
            if atoms_if_bond:
                core_atom = atoms_if_bond[0]
                h_core = atoms_if_bond[1]
                fragment_atom = atoms_if_bond[2]
                h_frag = atoms_if_bond[3]
            else:
                h_core = None
                h_frag = None
            try:
                print("PERFORMING INDIVIDUAL GROWING...")
                print("HYDROGEN ATOMS IN INSTRUCTIONS:  {}    {}".format(h_core, h_frag))
                grow_fragment(complex_pdb, fragment_pdb, core_atom, fragment_atom, iterations, criteria, plop_path, sch_python,
                     pele_dir, contrl, license, resfold, report, traject, pdbout, cpus, distcont, threshold, epsilon,
                     condition, metricweights, nclusters, pele_eq_steps, restart, min_overlap, max_overlap, ID, h_core,
                     h_frag, c_chain, f_chain, steps, temperature, seed, rotamers, banned, limit, mae, rename,
                     threshold_clash, steering, translation_high, rotation_high,
                     translation_low, rotation_low, explorative, radius_box, sampling_control, data, documents,
                     only_prepare, only_grow, no_check, debug, srun)
            except Exception:
                os.chdir(original_dir)
                traceback.print_exc()
                if debug: raise Exception()
        os.chdir(original_dir) 

if __name__ == '__main__':
    complex_pdb, iterations, criteria, plop_path, sch_python, pele_dir, \
    contrl, license, resfold, report, traject, pdbout, cpus, distcont, threshold, epsilon, condition, metricweights, \
    nclusters, pele_eq_steps, restart, min_overlap, max_overlap, serie_file, \
    c_chain, f_chain, steps, temperature, seed, rotamers, banned, limit, mae, \
    rename, threshold_clash, steering, translation_high, rotation_high, \
    translation_low, rotation_low, explorative, radius_box, sampling_control, data, documents, \
    only_prepare, only_grow, no_check, debug, protocol, test, srun = parse_arguments()
    
    main(complex_pdb, serie_file, iterations, criteria, plop_path, sch_python, pele_dir, contrl, license, resfold,
             report, traject, pdbout, cpus, distcont, threshold, epsilon, condition, metricweights,
             nclusters, pele_eq_steps, restart, min_overlap, max_overlap,
             c_chain, f_chain, steps, temperature, seed, rotamers, banned, limit, mae,
             rename, threshold_clash, steering, translation_high, rotation_high,
             translation_low, rotation_low, explorative, radius_box, sampling_control, data, documents,
             only_prepare, only_grow, no_check, debug, protocol, test, srun)

